#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **string joining** (i.e., concatenation of two or more strings,
typically separated by a delimiting substring) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type import types
from betse.util.type.types import type_check, IterableTypes

# ....................{ JOINERS                           }....................
def join(*texts) -> str:
    '''
    Concatenation of the passed strings with no separating delimiter.

    This is a convenience function wrapping the standard ``"".join((...))``
    method, whose syntax is arguably overly verbose.
    '''

    return join_on(*texts, delimiter='')


@type_check
def join_by_index(iterable: IterableTypes, subiterable_index: object) -> str:
    '''
    Concatenation of each string at the passed key or index of each subiterable
    of the passed iterable, with no separating delimiter.

    Parameters
    ----------
    iterable : IterableTypes
        Iterable of subiterables to be summed.
    subiterable_index : object
        Object with which to index each subiterable of this iterable. The type
        of this object *must* be a type accepted by the ``__getitem__()``
        special method of each subiterable. If each subiterable is a:

        * **Mapping** (e.g., :class:`dict`), this object *must* be hashable.
        * **Sequence** (e.g., :class:`list`, :class:`tuple`), this object
          *must* be either:

          * An integer.
          * A :func:`slice` object.

    Returns
    ----------
    str
        Concatenation of each string at this key or index of each subiterable
        of this iterable, with no separating delimiter.
    '''

    # Efficiency and simplicity combine here to form SuperHappyFunFunction.
    return ''.join(subiterable[subiterable_index] for subiterable in iterable)

# ....................{ JOINERS ~ on                      }....................
def join_on_newline(*texts) -> str:
    '''
    Join the passed strings with newline as the separating delimiter.

    This is a convnience function wrapping the standard
    ``"\n".join((...))`` method, whose syntax is arguably overly verbose.
    '''

    return join_on(*texts, delimiter='\n')


def join_on_dot(*texts) -> str:
    '''
    Join the passed strings with a period as the separating delimiter.

    This is a convnience function wrapping the standard
    ``".".join((...))`` method, whose syntax is arguably overly verbose.
    '''

    return join_on(*texts, delimiter='.')


def join_on_space(*texts) -> str:
    '''
    Join the passed strings with a space as the separating delimiter.

    This is a convnience function wrapping the standard
    ``" ".join((...))`` method, whose syntax is arguably overly verbose.
    '''

    return join_on(*texts, delimiter=' ')


@type_check
def join_on(*texts: IterableTypes, delimiter: str) -> str:
    '''
    Join the passed strings with the passed separating delimiter.

    This is a convenience function wrapping the standard
    ``"...".join((...))`` method, whose syntax is arguably overly obfuscated.

    Parameters
    ----------
    texts : SequenceTypes
        Tuple of all strings to be joined, consisting of either:

        * One or more strings.
        * One iterable of strings.
    delimiter : str
        Substring to join each such string on.

    Returns
    ----------
    str
        String joined from the passed strings with the passed delimiter.
    '''

    # If only one object was passed and this object is a non-string iterable
    # (e.g, list, tuple), default the list of passed strings to this object.
    if len(texts) == 1 and types.is_iterable_nonstr(texts[0]):
        texts = texts[0]

    # Join these strings.
    return delimiter.join(texts)

# ....................{ JOINERS ~ as                      }....................
@type_check
def join_as(
    *texts,
    delimiter_if_two: str,
    delimiter_if_three_or_more_nonlast: str,
    delimiter_if_three_or_more_last: str
) -> str:
    '''
    Join the passed strings in a human-readable manner.

    If:

    * No strings are passed, the empty string is returned.
    * One string is passed, this string is returned as is without modification.
    * Two strings are passed, these strings are joined with the passed
      ``delimiter_if_two`` separator.
    * Three or more strings are passed:

      * All such strings except the last two are joined with the passed
        ``delimiter_if_three_or_more_nonlast`` separator.
      * The last two such strings are joined with the passed
        ``delimiter_if_three_or_more_last`` separator.

    Parameters
    ----------
    texts : tuple[str]
        Tuple of all strings to be joined.
    delimiter_if_two : str
        String separating each item of ``texts`` if ``len(texts) == 2``.
    delimiter_if_three_or_more_nonlast : str
        String separating each item *except* the last two of ``texts`` if
        ``len(texts) >= 3``.
    delimiter_if_three_or_more_last : str
        String separating the last two items of ``texts`` if
        ``len(texts) >= 3``.

    Returns
    ----------
    str
        Resulting string as described above.

    Examples
    ----------
    >>> join_as(
    ...     ('Fulgrim', 'Perturabo', 'Angron', 'Mortarion'),
    ...     delimiter_if_two=' and ',
    ...     delimiter_if_three_or_more_nonlast=', ',
    ...     delimiter_if_three_or_more_last=', and '
    ... )
    'Fulgrim, Perturabo, Angron, and Mortarion'
    '''

    # Number of passed strings.
    texts_count = len(texts)

    # If no strings are passed, return the empty string.
    if texts_count == 0:
        return ''
    # If one string is passed, return this string as is.
    elif texts_count == 1:
        return texts[0]
    # If two strings are passed, return these strings joined appropriately.
    elif texts_count == 2:
        return '{}{}{}'.format(
            texts[0], delimiter_if_two, texts[1])
    # If three or more strings are passed, return these strings joined
    # appropriately.
    else:
        # All such strings except the last two, joined appropriately.
        texts_nonlast = join_on(
            *texts[0:-2], delimiter=delimiter_if_three_or_more_nonlast)

        # The last two such strings, joined appropriately.
        texts_last = '{}{}{}'.format(
            texts[-2], delimiter_if_three_or_more_last, texts[-1])

        # Return these two substrings, joined appropriately.
        return '{}{}{}'.format(
            texts_nonlast, delimiter_if_three_or_more_nonlast, texts_last)

# ....................{ JOINERS ~ as : conjunction        }....................
@type_check
def join_as_conjunction(*texts: str) -> str:
    '''
    Conjunctively join all passed strings in a human-readable manner.

    Specifically:

    * All passed strings excluding the last two are joined with ``, ``.
    * The last two passed strings are joined with ``, and ``.
    '''

    return join_as(
        *texts,
        delimiter_if_two=' and ',
        delimiter_if_three_or_more_nonlast=', ',
        delimiter_if_three_or_more_last=', and '
    )


@type_check
def join_as_conjunction_double_quoted(*texts: str) -> str:
    '''
    Conjunctively double-quote and join all passed strings in a human-readable
    manner.

    Specifically:

    * All passed strings are double-quoted.
    * All passed strings excluding the last two are joined with ``, ``.
    * The last two passed strings are joined with ``, and ``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Tuple of all passed strings double-quoted. Since the "*" operator applied
    # to this tuple below requires a sequence rather than generator, this is
    # the most space-efficient available sequence (i.e., frozen tuple).
    texts_quoted = tuple(strs.double_quote(text) for text in texts)

    # Conjunctively join these strings.
    return join_as_conjunction(*texts_quoted)


@type_check
def join_iterable_as_conjunction_double_quoted(iterable: IterableTypes) -> str:
    '''
    Conjunctively double-quote and join all string items of the passed iterable
    in a human-readable manner, implicitly converting each item of this
    iterable to a string as needed.

    See Also
    ----------
    :func:`join_as_conjunction_double_quoted`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable import iterables

    # Iterable of the same type as that of the passed iterable, implicitly
    # converting each item of this iterable to a string as needed.
    iterable_strs = iterables.to_iterable(iterable=iterable, item_cls=str)

    # Return these strings conjunctively double-quoted and joined.
    return join_as_conjunction_double_quoted(*iterable_strs)

# ....................{ JOINERS ~ as : conjunction        }....................
@type_check
def join_as_disjunction(*texts: str) -> str:
    '''
    Disjunctively join all passed strings in a human-readable manner.

    Specifically:

    * All passed strings excluding the last two are joined with ``, ``.
    * The last two passed strings are joined with ``, or ``.
    '''

    return join_as(
        *texts,
        delimiter_if_two=' or ',
        delimiter_if_three_or_more_nonlast=', ',
        delimiter_if_three_or_more_last=', or '
    )


@type_check
def join_as_disconjunction_double_quoted(*texts: str) -> str:
    '''
    Disjunctively double-quote and join all passed strings in a human-readable
    manner.

    Specifically:

    * All passed strings are double-quoted.
    * All passed strings excluding the last two are joined with ``, ``.
    * The last two passed strings are joined with ``, or ``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strs

    # Tuple of all passed strings double-quoted. See
    # join_as_conjunction_double_quoted().
    texts_quoted = tuple(strs.double_quote(text) for text in texts)

    # Disjunctively join these strings.
    return join_as_disjunction(*texts_quoted)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **mapping** (i.e., :class:`dict`-like types or instances)
functionality.
'''

# ....................{ IMPORTS                           }....................
from betse.exceptions import BetseMappingKeyException
from betse.util.type.types import (
    type_check, MappingType, HashableType, TestableOrNoneTypes)
from copy import deepcopy

# ....................{ GETTERS                           }....................
@type_check
def get_key_value(
    mapping: MappingType, key: HashableType, **kwargs) -> object:
    '''
    Value of the passed key in the passed mapping if this mapping contains this
    key *or* raise an exception otherwise (i.e., if this mapping contains no
    such key), optionally validated to be of the passed type.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be inspected.
    key : HashableType
        Key whose value is to be retrieved.

    All remaining keyword arguments are passed as is to the
    :func:`get_key_value_or_default` function.

    Returns
    ----------
    object
        Value of this key in this mapping.

    Raises
    ----------
    BetseMappingKeyException
        If this mapping contains no such key.
    BetseTypeException
        If the ``value_type`` parameter is non-``None`` and the type of the
        current value of this key is *not* an instance of ``value_type``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Value of this key in this mapping if any *OR* the sentinel otherwise.
    key_value = get_key_value_or_sentinel(
        mapping=mapping, key=key, **kwargs)

    # If this mapping contains no such key, raise an exception.
    if key_value is SENTINEL:
        raise BetseMappingKeyException(
            'Mapping key "{}" not found.'.format(key))
    # Else, this mapping contains this key.

    # Return this value.
    return key_value


@type_check
def get_key_value_or_sentinel(
    mapping: MappingType, key: HashableType, **kwargs) -> object:
    '''
    Value of the passed key in the passed mapping if this mapping contains this
    key *or* the sentinel singleton otherwise (i.e., if this mapping contains
    no such key), optionally validated to be of the passed type.

    This function enables callers to safely distinguish between non-existing
    keys and existing keys whose values are ``None``.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be inspected.
    key : HashableType
        Key whose value is to be retrieved.

    All remaining keyword arguments are passed as is to the
    :func:`get_key_value_or_default` function.

    Returns
    ----------
    object
        Either:

        * If this dictionary contains this key, this key's value.
        * Else, the **sentinel singleton** (i.e.,
          :attr:`betse.util.type.obj.sentinels.SENTINEL`).

    Raises
    ----------
    BetseTypeException
        If the ``value_type`` parameter is non-``None`` and the type of the
        current value of this key is *not* an instance of ``value_type``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj.sentinels import SENTINEL

    # Return the current value of this key in this mapping if any *OR* the
    # sentinel otherwise.
    return get_key_value_or_default(
        mapping=mapping, key=key, value_default=SENTINEL, **kwargs)


@type_check
def get_key_value_or_default(
    # Mandatory parameters.
    mapping: MappingType,
    key: HashableType,
    value_default: object,

    # Optional parameters.
    value_type: TestableOrNoneTypes = None,
) -> object:
    '''
    Value of the passed key in the passed mapping if this mapping contains this
    key *or* the passed default value otherwise (i.e., if this mapping contains
    no such key), optionally validated to be of the passed type.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be inspected.
    key : HashableType
        Key to return the current value of.
    value_default : object
        Default value to be returned if this dictionary contains no such key.
    value_type : TestableOrNoneTypes
        Expected type of the current value of this key. This function
        effectively performs the equivalent of the :meth:`type_check` decorator
        at runtime by raising an exception if all of the following apply:

        * This type is *not* ``None``.
        * This value is *not* this default value, implying this dictionary to
          contain this key.
        * This value is *not* an instance of this type.

        Defaults to ``None``, in which case no such type checking is performed.

    Returns
    ----------
    object
        Either:

        * If this dictionary contains this key, this key's value.
        * Else, this default value.

    Raises
    ----------
    BetseTypeException
        If the ``value_type`` parameter is non-``None`` and the type of the
        current value of this key is *not* an instance of ``value_type``.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objtest

    # Value of this key in this mapping if any *OR* this default value.
    key_value = mapping.get(key, value_default)

    # If this value is to be type-checked *AND* is *NOT* this default value
    # (which by definition already satisfies caller requirements regardless of
    # type), type-check this value.
    if value_type is not None and key_value is not value_default:
        objtest.die_unless_instance(obj=key_value, cls=value_type)

    # Return this value.
    return key_value

# ....................{ CONVERTERS                        }....................
@type_check
def to_str_flat(mapping: MappingType) -> str:
    '''
    Human-readable string flattening all key-value pairs of the passed
    dictionary.

    For readability, each such pair is sorted in ascending lexicographic order
    and formatted as ``{key}: {value}\n``. Ergo:

    * Each key and value is reduced to a string.
    * Each key is delimited from its value by a colon.
    * Each key-value pair is delimited by a newline.

    Caveats
    ----------
    **This function assumes all keys and values of this dictionary to be
    trivially reducible to terse strings,** where "terse" typically implies
    these strings to be no longer than a standard terminal width of 80
    characters. When this is *not* the case, the returned string is likely to
    be non-human-readable.

    Parameters
    ----------
    mapping: MappingType
        Dictionary to be flattened.

    Returns
    ----------
    str
        Human-readable string flattening all key-value pairs of this
        dictionary.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strjoin

    # Return the human-readable string produced by joining on newline a
    # generator comprehension yielding the colon-delimited name and value of
    # all environment variables sorted in lexicographic order.
    #
    # Note that the "environ" object is of non-standard type "os._Environ",
    # which the pprint.pformat() fails to recognize and hence format as a
    # "dict"-compatible mapping. Ergo, passing "environ" directly to the
    # iterables.to_str() function would produce a non-human-readable string.
    # While this can, of course, be ameliorated by converting "environ" to a
    # "dict" first (e.g., "iterables.to_str(dict(environ))"), doing so still
    # produces less human-readable output than the current approach.
    # return iterables.to_str(dict(environ))
    return strjoin.join_on_newline(
        '{}: {}'.format(key, mapping[key])
        for key in iter_keys_ascending(mapping)
    )

# ....................{ COPIERS                           }....................
@type_check
def copy_deep(mapping: MappingType) -> MappingType:
    '''
    Dictionary of all key-value pairs deeply (i.e., recursively) duplicated
    from the passed dictionary.

    This function should *always* be called in lieu of the standard
    :meth:`dict.__init__` and :meth:`dict.copy` methods, which only perform
    shallow dictionary copies. These copies fail to copy data structures nested
    in the values of the original dictionary, inviting subtle synchronization
    woes on subsequently modifying either the original or copied dictionaries.

    Parameters
    ----------
    mapping: MappingType
        Dictionary to be deeply copied.

    Returns
    ----------
    MappingType
        Dictionary of all key-value pairs deeply (i.e., recursively) duplicated
        from the passed dictionary.
    '''

    #FIXME: Does this simplistic approach guarantee the returned mapping to be
    #of the same type as the passed mapping?
    return deepcopy(mapping)


#FIXME: Well, this is rather awkward. Rather than define a completely separate
#function, it would be dramatically preferable to simply pass a new optional
#"keys_remove: IterableOrNoneTypes = None" parameter to the copy_deep()
#function defined above. When this parameter is:
#
#* "None", the copy_deep() function should reduce to its current one-liner.
#* Non-"None", the copy_deep() function should generalize to this function's
#  current implementation -- with the obvious caveat that the remove_key()
#  function should be generalized to accept an iterable of keys to be removed.
@type_check
def copy_map_sans_key(mapping: MappingType, key: HashableType) -> MappingType:
    '''
    Dictionary of all key-value pairs excluding that whose key is the passed
    key deeply (i.e., recursively) duplicated from the passed dictionary.

    Parameters
    ----------
    mapping: MappingType
        Dictionary to be deeply copied.
    key : HashableType
        Key to be removed from this dictionary.

    Returns
    ----------
    MappingType
        Dictionary of all key-value pairs excluding that whose key is this
        key deeply (i.e., recursively) duplicated from this dictionary.

    Raises
    ----------
    :class:`KeyError`
        If this dictionary contains no such key.

    See Also
    ----------
    :func:`copy_deep`
        Further details on map copying.
    :func:`remove_key`
        Further details on key removal.
    '''

    # Deep copy of this dictionary.
    mapping_copy = copy_deep(mapping)

    # Remove this key from this copy in-place.
    remove_key(mapping=mapping_copy, key=key)

    # Return this copy.
    return mapping_copy

# ....................{ INVERTERS                         }....................
#FIXME: Rename to simply invert_unique().
@type_check
def invert_map_unique(mapping: MappingType) -> MappingType:
    '''
    Dictionary inverted from the passed dictionary if no two key-value pairs of
    this dictionary share the same values *or* raise an exception otherwise.

    Specifically, the returned dictionary maps from each value to each key of
    the passed dictionary *and* is guaranteed to be the same type as that of
    the passed dictionary.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to be inverted. The type of this dictionary *must* define an
        ``__init__`` method accepting a single parameter whose value is an
        iterable of 2-iterables ``(key, value)`` providing all key-value pairs
        with which to initialize a new such dictionary. See the
        :meth:`dict.__init__` method for further details.

    Returns
    ----------
    MappingType
        Dictionary inverted from this dictionary as detailed above.

    Raises
    ----------
    BetseMappingException
        If one or more key-value pairs of this dictionary share the same
        values.

    See Also
    ----------
    https://stackoverflow.com/a/1679702/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable.mapping import maptest

    # If any values of this dictionary are are duplicates, raise an exception.
    maptest.die_unless_values_unique(mapping)

    # Type of this dictionary.
    mapping_type = type(mapping)

    # If this is an unordered dictionary, return a dictionary comprehension
    # efficiently inverting this dictionary in the trivial way.
    if mapping_type is dict:
        return {value: key for key, value in mapping.items()}
    # Else, this is possibly an ordered dictionary. In this case, a
    # considerably less trivial and slightly less efficient approach is needed.
    else:
        # Iterable of reversed 2-iterables "(value, pair)" for each key-value
        # pair of the passed dictionary. Dismantled, this is:
        #
        # * "mapping.items()", an iterable of 2-iterables "(key, value)" for
        #   each key-value pair of the passed dictionary.
        # * "reversed", a builtin which when passed such a 2-iterable returns
        #   the reversed 2-iterable "(value, pair)" for that key-value pair.
        # * "map(...)", a builtin applying the prior builtin to each such pair.
        value_key_pairs = map(reversed, mapping.items())

        # Return a new instance of this type of dictionary by invoking the
        # "dict(iterable)" form of this type's __init__() method. To quote the
        # dict.__init__() docstring:
        #
        # "dict(iterable) -> new dictionary initialized as if via:
        #      d = {}
        #      for k, v in iterable:
        #          d[k] = v"
        return mapping_type(value_key_pairs)

# ....................{ REMOVERS                          }....................
@type_check
def remove_key(mapping: MappingType, key: HashableType) -> None:
    '''
    Remove the key-value pair whose key is the passed key from the passed
    dictionary **in-place** (i.e., by modifying this dictionary rather than
    creating and returning a new dictionary with this key removed) if this
    dictionary contains this key *or* raise an exception otherwise.

    This function is a caller convenience improving codebase readability and
    efficiency. Although there exist multiple means of removing key-value pairs
    from dictionaries, this function implements the most efficient approach.
    These include:

    * The ``del mapping[key]`` idiom, known to be the most efficient approach.
    * The :meth:`dict.pop` method, known to be slightly less efficient than the
      idiomatic approach.

    Parameters
    ----------
    mapping : MappingType
        Dictionary to remove this key from.
    key : HashableType
        Key to be removed from this dictionary.

    Raises
    ----------
    :class:`KeyError`
        If this dictionary contains no such key.

    See Also
    ----------
    :func:`copy_map_sans_key`
        Function creating and returning a new dictionary with this key removed.
    '''

    # The best things in life are free.
    del mapping[key]

# ....................{ ITERATORS                         }....................
@type_check
def iter_keys_ascending(mapping: MappingType) -> list:
    '''
    Iterable of all keys of the passed dictionary sorted in ascending order.

    Parameters
    ----------
    mapping : MappingType
        Dictionary whose keys are to be sorted.

    Returns
    ----------
    list
        Iterable of all keys of this dictionary sorted in ascending order.

    See Also
    ----------
    :func:`betse.util.type.iterable.itersort.sort_ascending`
        Further details on ascending sorting.
    '''

    # Return a list of all keys of this dictionary sorted in ascending order.
    #
    # Note that the itersort.sort_ascending() function is intentionally *NOT*
    # called here, as that function currently fails to support dictionary views
    # in an efficient manner. Naturally, this should be rectified at some time.
    return sorted(mapping)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all YAML-backed file wrapper subclasses.
'''

# ....................{ IMPORTS                            }....................
from betse.lib.yaml import yamls
from betse.lib.yaml.abc.yamlabc import YamlABC
from betse.util.io.log import logs
from betse.util.path import dirs, pathnames
from betse.util.path.dirs import DirOverwritePolicy
from betse.util.type.descriptor.descs import abstractclassproperty_readonly
from betse.util.type.iterable.iterators import empty_iterator
from betse.util.type.types import type_check, IterableTypes

# ....................{ SUPERCLASSES                       }....................
class YamlFileABC(YamlABC):
    '''
    Abstract base class of all **YAML-backed file wrapper** (i.e., high-level
    object wrapping a low-level mapping or sequence of YAML-backed
    configuration settings both loadable from and savable back to a
    YAML-formatted file) subclasses.

    Caveats
    ----------
    The ``conf_filename`` parameter accepted by most methods of this class
    (e.g., :meth:`load`, :meth:`save`) should be suffixed by a valid YAML
    filetype -- namely, either ``.yml`` or ``.yaml``. This class does *not*
    erforce this recommendation but does log non-fatal warnings where violated.

    External callers should *never* access the low-level :meth:`conf` property,
    as doing so violates type safety and value constraints. The configuration
    settings persisted by this property are safely retrievable and modifiable
    *only* via :func:`yaml_alias` data descriptors defined by subclasses.

    Attributes
    ----------
    _conf_basename : StrOrNoneTypes
        Basename of :attr:`_conf_filename` if non-``None`` *or* ``None``
        otherwise.
    _conf_dirname : StrOrNoneTypes
        Absolute dirname of :attr:`_conf_filename` if non-``None`` *or*
        ``None`` otherwise.
    _conf_filename : StrOrNoneTypes
        Absolute filename of the low-level YAML-formatted configuration file
        from which this object was most recently deserialized if any *or*
        ``None`` otherwise.
    '''

    # ..................{ MAKERS                             }..................
    @classmethod
    @type_check
    def make(cls, conf_filename: str, *args, **kwargs) -> (
        'betse.lib.yaml.abc.yamlfileabc.YamlFileABC'):
        '''
        Create and return a YAML file wrapper of this subclass type,
        deserialized from the passed YAML-formatted file into a low-level
        mapping or sequence internally persisted in this wrapper.

        Parameters
        ----------
        conf_filename : str
            Absolute or relative filename of the source file to be
            deserialized.

        All other parameters are passed as is to the :meth:`__init__` method.

        Returns
        ----------
        YamlFileABC
            This configuration, returned for convenience chaining in callers.
        '''

        # Instance of this subclass, passed all passed variadic parameters.
        yaml_file_conf = cls(*args, **kwargs)

        # Deserialize the passed file into this instance.
        yaml_file_conf.load(conf_filename)

        # Return this instance.
        return yaml_file_conf

    # ..................{ INITIALIZERS                       }..................
    def __init__(self) -> None:
        '''
        Initialize this YAML file wrapper in the **unload state** (i.e.,
        associated with *no* low-level YAML-formatted configuration file).
        '''

        # Initialize our superclass.
        super().__init__()

        # Initialize this simulation configuration in the unload state. To
        # avoid extraneous logging, the unload() method body is repeated here.
        self._conf_basename = None
        self._conf_dirname = None
        self._conf_filename = None

    # ..................{ PROPERTIES ~ read-only             }..................
    # Read-only properties, preventing callers from resetting these attributes.

    @property
    def conf_basename(self) -> str:
        '''
        Basename of :attr:`conf_filename` if a prior call to the :meth:`load`
        method successfully loaded such a file *or* raise an exception
        otherwise.

        Raises
        ----------
        BetseYamlException
             If the :meth:`load` method has yet to be called.
        '''

        # If this property has yet to be set, raise an exception.
        self.die_unless_loaded()

        # Return this property.
        return self._conf_basename


    @property
    def conf_filename(self) -> str:
        '''
        Absolute filename of the low-level YAML-formatted file from which this
        file was deserialized if a prior call to the :meth:`load` method
        successfully loaded such a file *or* raise an exception otherwise.

        Raises
        ----------
        BetseYamlException
             If the :meth:`load` method has yet to be called.
        '''

        # If this property has yet to be set, raise an exception.
        self.die_unless_loaded()

        # Return this property.
        return self._conf_filename


    @property
    def conf_dirname(self) -> str:
        '''
        Absolute direname of :attr:`conf_filename` if a prior call to the
        :meth:`load` method successfully loaded such a file *or* raise an
        exception otherwise.

        Raises
        ----------
        BetseYamlException
             If the :meth:`load` method has yet to be called.
        '''

        # If this property has yet to be set, raise an exception.
        self.die_unless_loaded()

        # Return this property.
        return self._conf_dirname

    # ..................{ LOADERS                            }..................
    @type_check
    def load(self, conf_filename: str, **kwargs) -> None:
        '''
        Deserialize (i.e., load, read) the passed YAML-formatted file into a
        low-level mapping or sequence internally persisted in this wrapper.

        This method may be safely called multiple times with different files.
        This method implicitly calls the :meth:`unload` method if this is
        *not* the first call to this method, thus automatically deassociating
        this wrapper from all previously loaded files as needed.

        Parameters
        ----------
        conf_filename : str
            Absolute or relative filename of the source file to be
            deserialized.

        All remaining keyword arguments are passed as is to the
        :func:`betse.lib.yaml.yamls.load` function.
        '''

        # If a file is already loaded, unload this file for safety. While doing
        # so effectively reduces to a noop with respect to this superclass,
        # subclass implementations of the unload() method may elect to perform
        # critical functionality that *MUST* be respected here.
        if self.is_loaded:
            self.unload()

        # Log this load.
        logs.log_debug('Loading YAML file: %s', conf_filename)

        # Low-level dictionary deserialized from this file.
        conf = yamls.load(filename=conf_filename, **kwargs)

        # Load this dictionary into our superclass.
        super().load(conf=conf)

        # Associate this object with this file *AFTER* successfully
        # deserializing this file.
        self._set_conf_filename(conf_filename)


    def unload(self) -> None:

        # If this file is actually loaded, log this operation.
        if self.is_loaded:
            logs.log_debug('Closing YAML file: %s', self._conf_filename)

        # Unload our superclass.
        super().unload()

        # Nullify all instance variables for safety.
        self._conf_basename = None
        self._conf_dirname = None
        self._conf_filename = None

    # ..................{ SAVERS                             }..................
    #FIXME: For brevity, rename the following parameters:
    #
    #* "is_conf_file_overwritable" to "is_file_overwritable".
    #* "conf_subdir_overwrite_policy" to "subdir_overwrite_policy".
    @type_check
    def save(
        self,

        # Mandatory parameters.
        conf_filename: str,

        # Optional parameters.
        is_conf_file_overwritable: bool = False,
        conf_subdir_overwrite_policy: DirOverwritePolicy = (
            DirOverwritePolicy.SKIP_WITH_WARNING),
    ) -> None:
        '''
        Serialize (i.e., save, write) the low-level mapping or sequence
        internally persisted in this wrapper to the YAML-formatted file with
        the passed filename, copying *all* external resources internally
        referenced by this mapping or sequence into that file's directory.

        This method effectively implements the "Save As..." GUI metaphor.
        Specifically, this method (in order):

        #. Serializes this mapping or sequence to the file with this filename,
           optionally overwriting the existing contents of this file depending
           on the passed ``is_conf_file_overwritable`` parameter.
        #. Recursively copies all relative subdirectories internally referenced
           (and hence required) by this file from the directory of the current
           file associated with this wrapper into the directory of the passed
           file, optionally overwriting the existing contents of these
           subdirectories depending on the passed
           ``conf_subdir_overwrite_policy`` parameter.
        #. Associates this wrapper with this filename.

        Parameters
        ----------
        conf_filename : str
            Absolute or relative filename of the target file to be serialized.
        is_conf_file_overwritable : optional[bool]
            If this target file already exists *and* this boolean is:

            * ``True``, this target file is silently overwritten.
            * ``False``, an exception is raised.

            Defaults to ``False``.
        conf_subdir_overwrite_policy : DirOverwritePolicy
            **Subdirectory overwrite policy** (i.e., strategy for handling
            existing subdirectories to be overwritten by this save), where the
            subdirectories in question are all subdirectories of the directory
            of this target file. Defaults to
            :attr:`DirOverwritePolicy.SKIP_WITH_WARNING`, ignoring each target
            subdirectory that already exists with a non-fatal warning.

        Raises
        ----------
        BetseDirException
            If the passed ``conf_subdir_overwrite_policy`` parameter is
            :attr:`DirOverwritePolicy.HALT_WITH_EXCEPTION` *and* one or more
            subdirectories of the target directory already exist that are also
            subdirectories of the source directory.
        BetseFileException
            If the passed ``is_conf_file_overwritable`` parameter is ``False``
            *and* this target file already exists.
        '''

        # Log this save.
        logs.log_debug(
            'Saving YAML file: %s -> %s', self._conf_filename, conf_filename)

        # Serialize this mapping or sequence to this file.
        #
        # Note that the die_unless_loaded() method need *NOT* be called here,
        # as the "conf" property implicitly does so on our behalf.
        yamls.save(
            container=self.conf,
            filename=conf_filename,
            is_overwritable=is_conf_file_overwritable,
        )

        # Absolute dirnames of the directories containing the current file and
        # the passed file, canonicalized to permit comparison below.
        src_dirname = pathnames.canonicalize(self.conf_dirname)
        trg_dirname = pathnames.canonicalize(
            pathnames.get_dirname(conf_filename))

        # If these directories differ, recursively copy all relative
        # subdirectories internally referenced and hence required by this file.
        if src_dirname != trg_dirname:
            # For the basename of each such subdirectory...
            #
            # Note that the ideal solution of recursively copying this source
            # directory into the directory of this target file (e.g., via
            # "dirs.copy(src_dirname, pathnames.get_dirname(conf_filename))")
            # fails for the following subtle reasons:
            # * This target directory may be already exist, which dirs.copy()
            #   prohibits even when the directory is empty.
            # * This target configuration file basename may differ from that of
            #   this source configuration file, necessitating a subsequent call
            #   to file.move().
            for conf_subdirname in self._iter_conf_subdir_basenames():
                # Absolute dirname of the source subdirectory.
                src_subdirname = pathnames.join(src_dirname, conf_subdirname)

                # Recursively copy from the old into the new subdirectory.
                dirs.copy_dir_into_dir(
                    src_dirname=src_subdirname,
                    trg_dirname=trg_dirname,
                    overwrite_policy=conf_subdir_overwrite_policy,

                    # Ignore all empty ".gitignore" files in all subdirectories
                    # of this source subdirectory. These files serve only as
                    # placeholders instructing Git to track otherwise empty
                    # subdirectories. Preserving such files only invites end
                    # user confusion.
                    ignore_basename_globs=('.gitignore',),
                )

        # Associate this object with this file *AFTER* successfully copying to
        # this file and all external paths required by this file.
        self._set_conf_filename(conf_filename)


    def save_inplace(self) -> None:
        '''
        Serialize the low-level dictionary internally stored in this object to
        the current YAML-formatted simulation configuration file associated
        with this object, replacing the prior contents of this file.

        This method effectively implements the "Save" GUI metaphor.
        '''

        # Log this save.
        logs.log_debug('Resaving YAML file: %s', self._conf_filename)

        # Resave this dictionary to this file.
        yamls.save(
            container=self.conf,
            filename=self.conf_filename,
            is_overwritable=True,
        )

    # ..................{ COPIERS                            }..................
    def copy(
        self,
        src_conf_filename: str,
        trg_conf_filename: str,
        **kwargs
    ) -> None:
        '''
        Copy the source to target YAML-formatted file with the passed filenames
        and *all* external resources from the directory of the former into that
        of the latter.

        Specifically, this method (in order):

        #. Deassociates this wrapper from any previously loaded file.
        #. Loads the passed source file into this wrapper.
        #. Saves this source file to the passed target file.
        #. Associates this wrapper with the passed target file.

        Note that this method is currently syntactic sugar for the following
        equivalent operations, where ``p`` is the current object:

        #. Calling ``p.load(src_conf_filename)``.
        #. Calling ``p.save(trg_conf_filename, **kwargs)``.

        Parameters
        ----------
        src_conf_filename : str
            Absolute or relative filename of the source YAML-formatted file to
            safely copy from.
        trg_conf_filename : str
            Absolute or relative filename of the target YAML-formatted file to
            safely copy from.

        All remaining keyword arguments are passed to the :meth:`save` method.
        '''

        # Load this source file into this wrapper, implicitly deassociating
        # this wrapper from any previously loaded file.
        self.load(src_conf_filename)

        # Save this source file to this target file.
        self.save(trg_conf_filename, **kwargs)

    # ..................{ SETTERS                            }..................
    @type_check
    def _set_conf_filename(self, conf_filename: str) -> None:
        '''
        Set the absolute path of the YAML-formatted file associated with this
        configuration.

        Design
        ----------
        To prevent external callers from unsafely setting this path, this
        setter is intentionally implemented as an manual setter rather than a
        more preferable :meth:`conf_filename` property setter.
        '''

        # Unique absolute filename of this file assigned *BEFORE* this file's
        # parent directory, ensuring the latter is non-empty.
        self._conf_filename = pathnames.canonicalize(conf_filename)

        # Basename of this file.
        self._conf_basename = pathnames.get_basename(self._conf_filename)

        # Unique absolute dirname of the parent directory of this file.
        self._conf_dirname = pathnames.get_dirname(self._conf_filename)

    # ..................{ SUBCLASS ~ optional                }..................
    # Methods intended to be optionally overriden by subclasses.

    def _iter_conf_subdir_basenames(self) -> IterableTypes:
        '''
        Generator iteratively yielding the basename of each **requisite
        subdirectory** (i.e., subdirectory internally referenced and hence
        required by the YAML-formatted file currently associated with this
        configuration) of the directory containing this file.

        For safety, this generator excludes all direct subdirectories of this
        parent directory that are *not* internally referenced by this file.

        By default, this superclass method returns the empty generator.

        Yields
        ----------
        str
            Basename of each such subdirectory.
        '''

        # If no file has been read, raise an exception.
        self.die_unless_loaded()

        # Default to the empty iterator.
        return empty_iterator()

# ....................{ SUPERCLASSES ~ default             }....................
#FIXME: Is this genuinely required anymore? It would probably be saner to
#simply refactor all remaining calls to the copy_default() method to call the
#copy() method instead with the default simulation configuration file.
#
#The current approach is a bit too "magical." In particular, the "Parameter"
#subclass' implementation of the "conf_default_filename" property assumes
#the application metadata singleton to have already been initialized, inviting
#complications whenever either that property or the copy_default() method are
#called. Since explicit is better than implicit, our usage of properties here
#is inadvisable at best and bug-prone at worst.
class YamlFileDefaultABC(YamlFileABC):
    '''
    Abstract base class of all **YAML-backed defaultable file wrapper** (i.e.,
    high-level object wrapping a low-level mapping or sequence of YAML-backed
    configuration settings both loadable from and savable back to a
    YAML-formatted file, optionally defaulting to a subclass-specific default
    YAML-formatted file) subclasses.

    This superclass augments the :class:`YamlFileABC` superclass with support
    for loading from a subclass-specific default YAML-formatted file (e.g.,
    default simulation configuration file).
    '''

    # ..................{ PROPERTIES                         }..................
    # Subclasses are required to implement these abstract class properties.

    @abstractclassproperty_readonly
    def conf_default_filename(cls) -> str:
        '''
        Absolute filename of the low-level **default YAML-formatted file**
        (i.e., file specific to this subclass, typically containing sane
        defaults of general interest to a wide audience).
        '''

        pass

    # ..................{ COPIERS                            }..................
    @type_check
    def copy_default(self, trg_conf_filename: str, **kwargs) -> None:
        '''
        Copy the default YAML-formatted file specific to this subclass to the
        the passed filename *and* all external resources from the directory of
        the former into that of the latter.

        This method effectively implements the "New..." GUI metaphor.

        Parameters
        ----------
        trg_conf_filename : str
            Absolute or relative filename of the target YAML-formatted file to
            safely copy from.

        All remaining keyword arguments are passed to the :meth:`save` method.

        See Also
        ----------
        :meth:`copy`
            Further details.
        '''

        # Thus spake BETSEthustra.
        self.copy(
            src_conf_filename=self.conf_default_filename,
            trg_conf_filename=trg_conf_filename,
            **kwargs
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **mapping classes** (i.e., classes implementing dictionary-like
functionality, typically by subclassing the builtin :class:`dict` container
type or an analogue thereof).
'''

#FIXME: Abstract away our usage of "OrderedDict" throughout the codebase. Why?
#Under Python >= 3.7, the builtin "dict" type now preserves insertion order and
#hence is now explicitly ordered. See the official Python 3.7 release notes as
#well as the following Python thread between Guido and an acolyte:
#    https://mail.python.org/pipermail/python-dev/2017-December/151283.html
#Note that this guarantee was technically implemented by Python 3.6 but only as
#an "implementation detail." Since Python 3.7 first publicized this guarantee
#as a language requirement, only Python >= 3.7 *OR* CPython 3.6 can be
#guaranteed to preserve dictionary insertion order.
#
#Specifically, refactor the codebase as follows:
#
#* Define a new "betse.util.py.pyversion" submodule.
#* Shift the existing betse.util.py.get_version() function into this submodule.
#* Define the following new functions in this submodule:
#  * "def is_greater_than_or_equal_to(version: VersionTypes) -> bool",
#    returning true only if the version of the active Python interpreter is at
#    least that of the passed version.
#  * "@func_cached def is_3_6_or_newer() -> bool", implemented in terms of the
#    aforementioned is_greater_than_or_equal_to() function and returning true
#    only if the the active Python interpreter is Python >= 3.6.
#  * "@func_cached def is_3_7_or_newer() -> bool", implemented in terms of the
#    aforementioned is_greater_than_or_equal_to() function and returning true
#    only if the the active Python interpreter is Python >= 3.7.
#* Define a new "DictOrdered" global attribute of this module. Due to the
#  complexity of the conditionals governing this attribute, we probably want to
#  assign this attribute in a new _init() function of this module. Maybe? In
#  any case, the implementation should resemble in partial pseudocode:
#      DictOrdered = (
#          dict if (
#               pyversion.is_3_7_or_newer() or
#              (pyversion.is_3_6_or_newer() and pyimpl.is_cpython())
#          ) else collections.OrderedDict
#      )
#* Replace all existing references to the "collections.OrderedDict" type with
#  the above "DictOrdered" type.
#FIXME: Actually, BETSE only supports Python >= 3.8 now. Ergo, just
#unconditionally assume dictionaries to be ordered and drop all existing
#references to the obsolete (and possibly now even deprecated)
#"collections.OrderedDict" type.

# ....................{ IMPORTS                            }....................
from betse.exceptions import (
    BetseMappingException, BetseMethodUnimplementedException)
from betse.util.type import types
from betse.util.type.types import (
    type_check,
    CallableTypes,
    HashableType,
    IteratorType,
    MappingType,
    MappingMutableType,
)
from collections import OrderedDict

# ....................{ GLOBALS                            }....................
_DEFAULT_DICT_ID = 0
'''
Unique arbitrary identifier with which to uniquify the class name of the next
:func:`DefaultDict`-derived type.
'''

# ....................{ CLASSES ~ default                  }....................
#FIXME: Donate back to StackOverflow. The standard "defaultdict" class is
#sufficiently useless that numerous users would probably find this useful.
@type_check
def DefaultDict(
    #FIXME: For disambiguity, rename to "missing_key_maker".
    missing_key_value: CallableTypes,
    initial_mapping: MappingType = None,
) -> MappingType:
    '''
    Sane **default dictionary** (i.e., dictionary implicitly setting an unset
    key to the value returned by a caller-defined callable passed both this
    dictionary and that key).

    Motivation
    ----------
    The standard :class:`collections.defaultdict` class is sadly insane.
    Happily, this custom class is not. Specifically, that class:

    * Requires that the caller-defined callable accept *no* arguments.
    * Requires that the initial key-value pairs to seed this dictionary with be
      awkwardly defined as keyword arguments rather than key-value pairs.

    Parameters
    ----------
    missing_key_value : CallableTypes
        Callable (e.g., function, lambda, method) called to generate the default
        value for a "missing" (i.e., undefined) key on the first attempt to
        access that key, passed first this dictionary and then this key and
        returning this value. This callable should have a signature resembling:
        ``def missing_key_value(self: DefaultDict, missing_key: object) ->
        object``. Equivalently, this callable should have the exact same
        signature as that of the optional :meth:`dict.__missing__` method.
    initial_mapping : optional[MappingType]
        Non-default dictionary providing the initial key-value pairs of this
        default dictionary. Defaults to ``None``, in which case this
        default dictionary is initially empty.

    Returns
    ----------
    MappingType
        Default dictionary initialized as described above.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.cls import classes

    # Dynamically generated default dictionary class specific to this callable.
    default_dict_class = classes.define_class(
        class_name=_get_default_dict_class_name(),
        class_attr_name_to_value={'__missing__': missing_key_value,},
        base_classes=(dict,),
    )

    # Instantiate the first and only instance of this class.
    default_dict = default_dict_class()

    # If an initial dictionary was passed, add all of its key-value pairs.
    if initial_mapping is not None:
        default_dict.update(initial_mapping)

    # Return this instance.
    return default_dict

# ....................{ CLASSES ~ dynamic                 }....................
class DynamicValue(object):
    '''
    **Dynamic value** (i.e., qbject encapsulating a single variable gettable
    and settable via callables predefined at object initialization time).

    This object typically encapsulates a variable repeatedly reassigned to.
    Since the value of such a variable is inconstant, this variable is *not*
    reliably passable or referrable to without introducing desynchronization
    issues (e.g., the ``sim.cc_cells[sim.iNa]`` Numpy array of all
    cell-specific sodium ion concentrations, unreliably reassigned to each
    simulation time step). This object wraps this variable's access and
    modification, permitting this variable to effectively be reliably passed
    and referred to without introducing such issues.

    Attributes
    ----------
    get_value : CallableTypes
        Callable (e.g., function, lambda, method) dynamically getting this
        variable's value.
    set_value : CallableTypes
        Callable (e.g., function, lambda, method) dynamically setting this
        variable's value.
    '''

    # ..................{ SLOTS                             }..................
    # Tuple of the names of all instance attributes permitted in instances of
    # this class. This slightly improves the time efficiency of attribute
    # access (by anywhere from 5% to 10%) and dramatically improves the space
    # efficiency of object storage (by several orders of magnitude).
    __slots__ = ('get_value', 'set_value',)

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(
        self, get_value: CallableTypes, set_value: CallableTypes) -> None:
        '''
        Initialize this dynamic value.

        Parameters
        ----------
        get_value : CallableTypes
            Callable (e.g., function, lambda, method) dynamically getting this
            variable's value.
        set_value : CallableTypes
            Callable (e.g., function, lambda, method) dynamically setting this
            variable's value.
        '''

        self.get_value = get_value
        self.set_value = set_value


class DynamicValueDict(MappingMutableType):
    '''
    Dictionary whose values are all **dynamic** (i.e., instances of the
    `DynamicValue` class, whose actual underlying values are gettable and
    settable via callables specific to those instances).

    Caveats
    ----------
    **All dictionary keys are predefined at dictionary initialization time.**
    Unlike standard dictionaries, this dictionary implementation does *not*
    permit additional key-value pairs to be added to or existing key-value
    pairs to be removed from this dictionary after initialization. While the
    value to which any key maps is arbitrarily modifiable, no key itself is
    modifiable.

    Attributes
    ----------
    _key_to_dynamic_value : MappingType
        Dictionary such that each:
        * Key is a key of the outer dictionary.
        * Value is the `DynamicValue` instance permitting that key's underlying
          variable to be dynamically get and set.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, key_to_dynamic_value: MappingType) -> None:
        '''
        Initialize this dictionary.

        Parameters
        ----------
        key_to_dynamic_value : MappingType
            Dictionary such that each:
            * Key is a key of the outer dictionary.
            * Value is the `DynamicValue` instance permitting that key's
              underlying variable to be dynamically get and set.
        '''

        #FIXME: Premature optimization. Reduce this to the following, please:
        #    from betse.util.type.iterable import iterables
        #    iterables.die_unless_items_instance_of(
        #        iterable=key_to_dynamic_value.values(), cls=DynamicValue)

        # If optimization is disabled, validate all values of the passed
        # dictionary to be instances of the "DynamicValue" class.
        if __debug__:
            for dynamic_value in key_to_dynamic_value.values():
                assert isinstance(dynamic_value, DynamicValue), (
                    '"{}" not a dynamic value.'.format(
                        types.trim(dynamic_value)))

        # Classify the passed parameters.
        self._key_to_dynamic_value = key_to_dynamic_value


    # ..................{ MAGIC ~ dict                      }..................
    def __getitem__(self, key: HashableType) -> object:
        '''
        Get the underlying value of the variable associated with the passed
        key.

        Parameters
        ----------
        key : HashableType
            Key to get the value of.

        Returns
        ----------
        object
            Underlying value of the variable associated with this key.

        Raises
        ----------
        KeyError
            If this key is *not* a key with which this dictionary was
            initialized.
        '''

        return self._key_to_dynamic_value[key].get_value()


    def __setitem__(self, key: HashableType, value: object) -> None:
        '''
        Set the underlying value of the variable associated with the passed
        key.

        Parameters
        ----------
        key : HashableType
            Key to set the value of.
        value : object
            Underlying value of the variable associated with this key to be
            set.

        Raises
        ----------
        KeyError
            If this key is *not* a key with which this dictionary was
            initialized.
        '''

        # Attempt to get this key *BEFORE* setting this key, implicitly raising
        # a "KeyError" exception if this key was not defined at initialization.
        # self._key_to_dynamic_value[key]

        # Set this key.
        self._key_to_dynamic_value[key].set_value(value)


    def __delitem__(self, key: HashableType) -> None:
        '''
        Unconditionally raise an exception.

        Since the size of this dictionary is invariant, deleting key-value
        pairs from this dictionary is strictly prohibited.

        Raises
        ----------
        BetseMethodUnimplementedException
            If this key is *not* a key with which this dictionary was
            initialized.
        '''

        raise BetseMethodUnimplementedException()

    # ..................{ MAGIC ~ container                 }..................
    def __len__(self) -> int:
        '''
        Get the size of this dictionary, guaranteed to be equal to the size of
        the dictionary with which this dictionary was initialized.

        Returns
        ----------
        int
            This size.
        '''

        return len(self._key_to_dynamic_value)

    # ..................{ MAGIC ~ iterable                  }..................
    def __iter__(self) -> IteratorType:
        '''
        Get an iterator over all keys with which this dictionary was
        initalized.

        Returns
        ----------
        IteratorType
            This iterator.
        '''

        return iter(self._key_to_dynamic_value)

# ....................{ CLASSES ~ ordered                 }....................
class OrderedArgsDict(OrderedDict):
    '''
    Ordered dictionary initialized by a sequence of key-value pairs.

    Motivation
    ----------
    This :class:`OrderedDict` subclass simplifies the definition of ordered
    dictionaries. As is common practice throughout the stdlib, the standard
    :class:`OrderedDict` class is defined by a sequence of 2-sequences
    ``(key, value)`` (e.g., tuple of 2-tuples) where each ``key`` and ``value``
    comprises one key-value pair of this dictionary.

    This :class:`OrderedDict` subclass is instead defined by a flat sequence of
    keys and values rather than a nested sequence of 2-sequences. In all other
    respects, this subclass is identical to the :class:`OrderedDict` class.

    Examples
    ----------
    >>> from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict
    >>> from collections import OrderedDict
    >>> tuatha_de_danann = OrderedArgsDict(
    ...     'Nuada', 'Nodens',
    ...     'Lugh', 'Lugus',
    ...     'Brigit', 'Brigantia',
    ... )
    >>> tuath_de = OrderedDict((
    ...     ('Nuada', 'Nodens'),
    ...     ('Lugh', 'Lugus'),
    ...     ('Brigit', 'Brigantia'),
    ... ))
    >>> for irish_name, celtic_name in tuatha_de_danann.items():
    ...     print('From Irish {} to Celtic {}.'.format(irish_name, celtic_name))
    From Irish Nuada to Celtic Nodens.
    From Irish Lugh to Celtic Lugus.
    From Irish Brigit to Celtic Brigantia.
    >>> tuatha_de_danann == tuath_de
    True
    '''


    def __init__(self, *key_value_pairs) -> None:
        '''
        Initialize this ordered dictionary with the passed tuple of sequential
        keys and values of even length.

        Attributes
        ----------
        key_value_pairs : tuple
            Tuple of sequential keys and values to initialize this ordered
            dictionary with. Each item of this tuple with:

            * Even index (e.g., the first and third items) defines a new key of
              this dictionary, beginning a new key-value pair.
            * Odd index (e.g., the second and fourth items) defines the value
              for the key defined by the preceding item, finalizing this
              existing key-value pair.

        Raises
        ----------
        :exc:`betse.exceptions.BetseMappingException`
             If this tuple is *not* of even length.
        '''

        # Avoid circular import dependencies.
        from betse.util.type.numeric import ints

        # If the passed tuple of key-value pairs is odd and hence omitted the
        # value for the final key, raise an exception.
        if ints.is_odd(len(key_value_pairs)):
            raise BetseMappingException(
                'Expected even number of key-value parameters, '
                'but received {} such parameters.'.format(
                    len(key_value_pairs)))

        # Zip object yielding 2-tuple key-value pairs of the format required by
        # the superclass, converted from this flat sequence of keys and values.
        key_value_pairs_nested = zip(
            key_value_pairs[0::2],
            key_value_pairs[1::2])

        # Initialize the superclass with this zip.
        super().__init__(key_value_pairs_nested)

# ....................{ CLASSES ~ reversible              }....................
#FIXME: Define an analogous "OneToOneDict" subclass of the "dict" superclass,
#mandating that no values ambiguously map to by two or more keys of the
#original dictionary. In theory, the implementation should be even more
#concise.
#FIXME: Unit test us up.
class ReversibleDict(dict):
    '''
    Bidirectional dictionary.

    This dictionary subclass provides a public :attr:``reverse`` instance
    variable, whose value is a dictionary providing both safe and efficient
    **reverse lookup** (i.e., lookup by values rather than keys) of the
    key-value pairs of the original dictionary.

    This dictionary explicitly supports reverse lookup of values ambiguously
    mapped to by two or more keys of the original dictionary. To disambiguate
    between these mappings, each key of the :attr:``reverse`` dictionary
    (signifying a value of the original dictionary) maps to a sequence
    containing all keys of the original dictionary mapping to that value.

    Caveats
    ----------
    **The :attr:``reverse`` dictionary is not safely modifiable.**

    Attempting to do so breaks the contractual semantics of this class in an
    unsupported manner. For safety, this dictionary is silently synchronized
    with *all* modifications to the original dictionary -- including additions
    of new keys, modifications of the values of existing keys, and deletions of
    existing keys. However, the reverse is *not* the case; the original
    dictionary is *not* silently synchronized with any modifications to the
    :attr:``reverse`` dictionary. (Doing so is technically feasible but beyond
    the scope of current use cases.)

    Attributes
    ----------
    reverse : dict
        Dictionary such that each:

        * Key is a value of the original dictionary.
        * Value is a list of all keys of the original dictionary mapping to
          that value of the original dictionary.

    See Also
    ----------
    https://stackoverflow.com/a/21894086/2809027
        Stackoverflow answer inspiring this class. Cheers, Basj!
    '''


    def __init__(self, *args, **kwargs) -> None:
        # Initialize the original dictionary.
        super().__init__(*args, **kwargs)

        # Initialize the reversed dictionary. For each key-value pair with
        # which the original dictionary is initialized, map this value to a
        # list of all corresponding keys for reverse lookup.
        self.reverse = {}
        for key, value in self.items():
            self.reverse.setdefault(value, []).append(key)


    def __setitem__(self, key: HashableType, value: object) -> None:
        # Map this key to this value in the original dictionary.
        super().__setitem__(key, value)

        # Map this value to this key in the reversed dictionary.
        self.reverse.setdefault(value, []).append(key)


    def __delitem__(self, key: HashableType) -> None:
        # Value to be deleted from the reversed dictionary.
        value = self[key]

        # Remove this key-value pair from the original dictionary.
        super().__delitem__(key)

        # Remove this value-key pair from the reversed dictionary.
        self.reverse[value].remove(key)

        # If this key is the last key mapping to this value in the original
        # dictionary, remove this value entirely from the reversed dictionary.
        # While arguably optional, doing so reduces space costs with no
        # concomitant increase in time costs.
        if not self.reverse[value]:
            del self.reverse[value]

# ....................{ PRIVATE                           }....................
def _get_default_dict_class_name() -> str:
    '''
    Name of the class of the next data descriptor created and returned by the
    next call to the :func:`DefaultDict` function, guaranteed to be unique
    across all such classes.

    Since the human-readability of this name is neither required nor desired,
    this is a non-human-readable name efficiently constructed from a private
    prefix and a unique arbitrary identifier.
    '''

    # Global variable modified below.
    global _DEFAULT_DICT_ID

    # Increment this identifier to preserve uniqueness.
    _DEFAULT_DICT_ID += 1

    # Return a unique classname suffixed by this identifier.
    return '__DefaultDict' + str(_DEFAULT_DICT_ID)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **Python identifier** (i.e., class, module, or attribute name)
facilities.
'''

# ....................{ TODO                              }....................
#FIXME: *WOOPS.* Hilarliously, the builtin "str" type already provides a
#standard method for efficiently detecting Python identifiers:
#    >>> str.isidentifier?
#    Signature: str.isidentifier(self, /)
#    Docstring:
#    Return True if the string is a valid Python identifier, False otherwise.
#    Use keyword.iskeyword() to test for reserved identifiers such as "def" and
#    "class".
#    Type:      method_descriptor
#Given that, the is_unqualified() function should be refactored as follows:
#    @type_check
#    def is_unqualified(text: str) -> bool:
#        return text.isidentifier()
#After doing so, remove the "_IDENTIFIER_UNQUALIFIED_REGEX_COMPILED" but *NOT*"
#"_IDENTIFIER_UNQUALIFIED_REGEX" globals, as the former is no longer required.
#FIXME: *WOOPS.* It's a good thing that str.isidentifier() exists, as PEP 3131
#introduces Unicode identifiers to Python:
#    https://www.python.org/dev/peps/pep-3131
#We never knew this was a thing, honestly. Unlike Raku (Perl 6), which openly
#embraces and arguably abuses Unicode for everything, we've never actually seen
#real-world Python code in the wild using Unicode identifiers. In response,
#we'll need to significantly update the regular expressions below. Or perhaps
#we simply don't care enough at the moment to generalize this....? *sigh*

# ....................{ IMPORTS                           }....................
import re
from betse.exceptions import BetsePyIdentifierException
from betse.util.type.types import type_check, RegexMatchType

# ....................{ CLASSES                           }....................
IDENTIFIER_ALPHANUMERIC_CHAR_CLASS = r'a-zA-Z0-9'
'''
Character class (excluding ``[`` and ``]`` delimiters) matching any character
of a **Python identifier** (i.e., class, module, or attribute name), excluding
the underscore character.

Of necessity, this character class is equivalent to the character class of all
alphanumeric ASCII characters.
'''


IDENTIFIER_CHAR_CLASS = IDENTIFIER_ALPHANUMERIC_CHAR_CLASS + r'_'
'''
Character class (excluding ``[`` and ``]`` delimiters) matching any character
of a **Python identifier** (i.e., class, module, or attribute name).
'''

# ....................{ REGEXES                           }....................
IDENTIFIER_CHAR_CAMELCASE_REGEX = (
    r'(?:(?<=[a-z0-9])|(?!^))([A-Z])(?=[a-z]|$)')
'''
Uncompiled regular expression matching the next first character of a contiguous
run of uppercase characters in a CamelCase-formatted Python identifier (e.g.,
the `I` in `Capitel IV`), excluding the first character of such identifier..

Examples
----------

This expression is intended to be used in substitutions converting CamelCase to
some other format. For example, to convert CamelCase to snake_case:

    >>> from betse.util.type.text import regexes
    >>> regexes.replace_substrs(
    ...     'MesseIoXaVIaX',
    ...     rexeges.IDENTIFIER_CHAR_CAMELCASE_REGEX,
    ...     r'_\1')
    Messe_io_xa_via_x
'''


IDENTIFIER_UNQUALIFIED_REGEX = r'[{}]+'.format(IDENTIFIER_CHAR_CLASS)
'''
Uncompiled regular expression matching an **unqualified Python identifier**
(i.e., class, module, or attribute name *not* prefixed by a package or module
name).
'''


IDENTIFIER_QUALIFIED_REGEX = (
    r'(?:{identifier_unqualified}\.)*{identifier_unqualified}'.format(
        identifier_unqualified=IDENTIFIER_UNQUALIFIED_REGEX))
'''
Uncompiled regular expression matching a **qualified Python identifier** (i.e.,
class, module, or attribute name possibly prefixed by a package or module
name).
'''

# ....................{ REGEX ~ compiled                  }....................
# These regular expressions are frequently referenced at application startup
# and thus unconditionally compiled. Specifically, this regex is referenced by
# the sanitize_camelcase() function called by the var_alias() descriptor
# repeatedly instantiated at class scope by "Parameters"-related classes.

_IDENTIFIER_UNQUALIFIED_REGEX_COMPILED = re.compile(
    IDENTIFIER_UNQUALIFIED_REGEX)
'''
Compiled regular expression matching an **unqualified Python identifier**
(i.e., class, module, or attribute name *not* prefixed by a package or module
name).
'''


_IDENTIFIER_SANITIZE_CAMELCASE_REGEX_COMPILED = re.compile(
    r'(?:^|[^a-zA-Z0-9]+)([a-z])?')
'''
Compiled regular expression matching either the string start *or* one or more
alphanumeric ASCII characters optionally followed by a lowercase alphabetic
ASCII character, internally required by the :func:`sanitize_camelcase`
function.
'''

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_unqualified(*texts: str) -> None:
    '''
    Raise an exception unless all passed strings are syntactically valid
    **unqualified Python identifiers** (i.e., class, module, or attribute names
    *not* prefixed by package or module names).

    See Also
    ----------
    :func:`is_unqualified`
        Further details.
    '''

    # For each such string, if this string is *NOT* a valid variable name,
    # raise an exception.
    for text in texts:
        if not is_unqualified(text):
            raise BetsePyIdentifierException(
                'String "{}" not a '
                'valid unqualified Python identifier.'.format(text))

# ....................{ TESTERS                           }....................
#FIXME: The current implementation is rather unspecial. Instead, we should:
#
#* Create a new "_IDENTIFIER_SPECIAL_REGEX_COMPILED" global whose value is a
#  compiled regular expression matching all syntactically valid special Python
#  identifiers.
#* Refactor this function to match the passed string against this global.
@type_check
def is_special(text: str) -> bool:
    '''
    ``True`` only if the passed string is a possibly syntactically invalid
    **special Python identifier** (i.e., class, module, or attribute name
    both prefixed *and* suffixed by the ``__`` delimiter).

    Caveats
    ----------
    **This function does not test whether this string is a syntactically valid
    Python identifier.** For efficiency, this function only tests whether this
    string is special.
    '''

    return text.startswith('__') and text.endswith('__')


@type_check
def is_unqualified(text: str) -> bool:
    '''
    ``True`` only if the passed string is a syntactically valid
    **unqualified Python identifier** (i.e., class, module, or attribute name
    *not* prefixed by a package or module name).
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Return True only if this string matches the desired regular expression.
    return regexes.is_match(
        text=text, regex=_IDENTIFIER_UNQUALIFIED_REGEX_COMPILED)

# ....................{ CONVERTERS ~ camel                }....................
@type_check
def convert_camelcase_to_snakecase(identifier: str) -> str:
    '''
    Convert the passed CamelCase-formatted Python identifier to snake_case
    (e.g., from ``ThePMRC`` to ``the_pmrc``).
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Munge this identifier.
    return regexes.replace_substrs(
        text=identifier,
        regex=IDENTIFIER_CHAR_CAMELCASE_REGEX,
        replacement=r'_\1',
    ).lower()


@type_check
def convert_camelcase_to_whitespaced_lowercase(identifier: str) -> str:
    '''
    Convert the passed CamelCase-formatted Python identifier to whitespaced
    lowercase (e.g., from ``CleanseIII`` to ``cleanse iii``).
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Munge this identifier.
    return regexes.replace_substrs(
        text=identifier,
        regex=IDENTIFIER_CHAR_CAMELCASE_REGEX,
        replacement=r' \1',
    ).lower()

# ....................{ SANITIZERS ~ camelcase            }....................
@type_check
def sanitize_camelcase(identifier: str) -> str:
    '''
    Sanitize the passed string into a valid Python identifier in CamelCase
    format derived from this string (e.g., from ``sim-grn`` to ``SimGrn``).

    Specifically, this function replaces each substring of one or more
    alphanumeric ASCII characters with:

    * If this substring is followed by a lowercase alphabetic ASCII
      character, that character uppercased (e.g., from ``['spider jerusalem']``
      to ``SpiderJerusalem``).
    * Else, the empty string (e.g., from ``['']`` to ````).

    Parameters
    ----------
    identifier : str
        String to be converted into a valid CamelCase-style Python identifier.

    Returns
    ----------
    str
        Valid CamelCase-style Python identifier converted from this string.

    Raises
    ----------
    BetsePyIdentifierException
        If the passed string contains no alphanumeric characters, in which case
        the resulting identifier would be empty and hence invalid.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Identifier sanitized from this unsanitized identifier as detailed above.
    identifier_sanitized = regexes.replace_substrs(
        text=identifier,
        regex=_IDENTIFIER_SANITIZE_CAMELCASE_REGEX_COMPILED,
        replacement=_sanitize_camelcase_match,
    )

    # Return this sanitized identifier is empty, this unsanitized identifier
    # must necessarily have contained no alphanumeric characters. In this case,
    # raise an exception.
    if not identifier_sanitized:
        raise BetsePyIdentifierException(
            'Unsanitized identifier "{}" contains no '
            'alphanumeric characters.'.format(identifier))

    # Return this sanitized identifier.
    return identifier_sanitized


def _sanitize_camelcase_match(match: RegexMatchType) -> str:
    '''
    Uppercase the lowercase alphabetic ASCII character optionally grouped by
    the passed regular expression match object if any *or* return the empty
    string otherwise.

    Parameters
    ----------
    match: RegexMatchType
        Match object optionally grouping a lowercase alphabetic ASCII
        character.

    Returns
    ----------
    str
        Either:

        * If this match object grouped a lowercase alphabetic ASCII character,
          this character uppercased.
        * Else, the empty string.
    '''

    return (match.group(1) or '').upper()

# ....................{ SANITIZERS ~ snakecase            }....................
#FIXME: Generalize to sanitize arbitrary strings. See the general-purpose
#sanitize_camelcase() function for inspiration.
@type_check
def sanitize_snakecase(identifier: str) -> str:
    '''
    Sanitize the passed string into a valid Python identifier in snake_case
    format derived from this string (e.g., from ``sim-gnr`` to ``sim_gnr``).

    Specifically, this function:

    * Replaces each of the following characters in this string with an
      underscore:
      * A hyphen.

    Parameters
    ----------
    identifier : str
        String to be converted into a valid snake_case-style Python identifier.

    Returns
    ----------
    str
        Valid snake_case-style Python identifier converted from this string.
    '''

    return identifier.replace('-', '_')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level tissue and event handling functionality.
'''

# ....................{ IMPORTS                           }....................
import copy
import numpy as np
from betse.exceptions import BetseSimTissueException
from betse.science.enum.enumconf import CellsPickerType, SolverType
from betse.science.math import modulate as mod
from betse.science.math import toolbox as tb
from betse.science.phase.phasecls import SimPhase
from betse.science.tissue.tisprofile import CutProfile, TissueProfile
from betse.science.tissue.event.tisevecut import SimEventCut
from betse.science.tissue.event.tisevevolt import SimEventPulseVoltage
from betse.science.tissue.picker.tispickcls import (
    TissuePickerABC,
    TissuePickerAll,
    TissuePickerColor,
    TissuePickerIndices,
    TissuePickerPercent,
)
from betse.science.tissue.picker.tispickimage import TissuePickerImage
from betse.util.io.log import logs
# from betse.util.type import types
from betse.util.type.types import type_check
from collections import OrderedDict
from random import shuffle

# ....................{ CLASSES                           }....................
#FIXME: Since this class handles both tissue profiles *AND* scheduled
#interventions, consider renaming this class to a less ambiguous name -- say, to
#"TissueEventHandler".

#FIXME: Document all instance variables defined within this class.

#FIXME: Include other channels in morphogen (dye) dynamics.
class TissueHandler(object):
    '''
    High-level tissue handler, managing all tissue-centric functionality
    including both tissue and cut profiles *and* scheduled interventions.

    This handler governs all:

    * Tissue profiles and objects required by these profiles, including all
      geometry-specifying bitmaps.
    * Scheduled interventions, even those *not* pertaining to tissue profiles
      (e.g., global scheduled interventions).

    Attributes (Profile: Tissue)
    ----------
    cell_target_inds : dict
        Dictionary mapping from the name of each tissue profile enabled by the
        current simulation configuration to a one-dimensional array of the
        indices of all cells in the cluster belonging to this tissue.
    tissue_default : TissueProfile
        **Default tissue profile** (i.e., profile applicable to all cells *not*
        already assigned to a non-default tissue profile). For convenience,
        this profile is accessible via both this instance variable *and* as a
        standard tissue profile (i.e., as a value of a key-value pair of the
        ordered :attr:`tissue_name_to_profile` dictionary whose key is the name
        of this profile).
    tissue_name_to_profile : OrderedDict
        Ordered dictionary mapping from the name of each tissue profile enabled
        by the current simulation configuration (in YAML list order) to the
        :class:`TissueProfile` instance encapsulating this profile.
    tissue_target_inds : dict
        Dictionary mapping from the name of each tissue profile enabled by the
        current simulation configuration to a one-dimensional array of the
        indices of all cell membranes in the cluster belonging to this tissue.

    Attributes (Profile: Cut)
    ----------
    cut_name_to_profile : OrderedDict
        Ordered dictionary mapping from the name of each cut profile enabled by
        the current simulation configuration (in YAML list order) to the
        :class:`CutProfile` instance encapsulating this profile.
    _wound_channel : {TRP, NoneType}
        Wound-induced TRP-based channel if the following conditions are all
        satisfied *or* ``None`` otherwise:

        * The current simulation configuration has enabled simulation of the
          mechanosensitive Na channel associated with wounding.
        * The cutting event has already been performed, thus generating a wound
          and activating this channel.

    Attributes (Event)
    ----------
    _event_voltage : {SimEventPulseVoltage, NoneType}
        **Voltage event** (i.e., event applying a directed voltage to the
        environmental boundary for a range of simulation time steps) if enabled
        by the current simulation configuration *or* ``None`` otherwise.
    event_cut : {SimEventCut, NoneType}
        **Cutting event** (i.e., event removing a region of the current cluster
        at some time step during the simulation phase) if enabled by the
        current simulation configuration *or* ``None`` otherwise.
    '''

    # ..................{ INITIALIZERS                      }..................
    # Avoid circular import dependencies.
    @type_check
    def __init__(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Initialize this tissue handler.

        Parameters
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        # Nullify all instance variables for safety.
        self.event_cut = None
        self._event_voltage = None
        self._wound_channel = None

        # Initialize tissue and cut profiles.
        self._init_tissues(p)
        self._init_cuts(p)

        # Initialize events *AFTER* tissue and cut profiles, as the former
        # frequently require the latter.
        self._init_events(p)


    @type_check
    def _init_tissues(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Internally initialize all tissue profiles defined by the passed
        simulation configuration.

        Specifically, this method encapsulates each low-level YAML-formatted
        list item defining a tissue profile in this configuration with a
        high-level instance of the :class:`CellsProfileABC` class.

        Parameters
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        # Ordered dictionaries mapping from the name of each each tissue profile
        # to the corresponding profile.
        self.tissue_name_to_profile = OrderedDict()

        # Nullify all remaining tissue-centric attributes for safety.
        self.cell_target_inds = None
        self.env_target_inds = None
        self.tissue_target_inds = None

        # If tissue profiles are disabled, silently noop.
        if not p.is_tissue_profiles:
            return

        # 1-based index of the default tissue profile.
        tissue_z_order = 1

        # Object assigning a cell cluster region to the default tissue profile.
        tissue_picker = TissuePickerAll()

        #FIXME: Set this picker for the "tissue_default" object, while still
        #setting the above tissue picker for the
        #"tissue_name_to_profile[p.tissue_default.name]" object -- implying
        #that these two objects should be different objects. Viola!
        # tissue_picker = TissuePickerImage(
        #     filename=p.tissue_default.picker_image_filename,
        #     dirname=p.conf_dirname)

        # For convenience, encapsulate the default tissue profile as a
        # high-level object accessible both as a separate instance variable
        # *AND* as a standard tissue profile (i.e., as a value of a key-value
        # pair whose key is the name of the default tissue profile).
        self.tissue_default = self.tissue_name_to_profile[
            p.tissue_default.name] = TissueProfile(
                name=p.tissue_default.name,
                z_order=tissue_z_order,
                picker=tissue_picker,

                # By definition, the cell cluster shares no gap junctions with
                # cells in the environment -- since, of course, there *ARE* no
                # cells in the environment.
                is_gj_insular=True,

                Dm_Na=p.tissue_default.Dm_Na,
                Dm_K=p.tissue_default.Dm_K,
                Dm_Cl=p.tissue_default.Dm_Cl,
                Dm_Ca=p.tissue_default.Dm_Ca,
                Dm_M=p.tissue_default.Dm_M,
                Dm_P=p.tissue_default.Dm_P,
            )

        # For each non-default tissue profile...
        for tissue_profile in p.tissue_profiles:
            # If a prior profile collides with this profile's name, this
            # profile is non-unique. In this case, raise an exception.
            if tissue_profile.name in self.tissue_name_to_profile:
                raise BetseSimTissueException(
                    'Tissue profile "{0}" non-unique '
                    '(i.e., two or more tissue profiles named "{0}").'.format(
                        tissue_profile.name))

            # 1-based index of this tissue profile.
            tissue_z_order += 1

            # Picker assigning a cell cluster region to this tissue profile.
            tissue_picker = None

            # Conditionally define this picker.
            if tissue_profile.picker_type is CellsPickerType.ALL:
                tissue_picker = TissuePickerAll()
            elif tissue_profile.picker_type is CellsPickerType.COLOR:
                tissue_picker = TissuePickerColor(
                    cells_color=tissue_profile.picker_cells_color)
            elif tissue_profile.picker_type is CellsPickerType.IMAGE:
                tissue_picker = TissuePickerImage(
                    filename=tissue_profile.picker_image_filename,
                    dirname=p.conf_dirname)
            elif tissue_profile.picker_type is CellsPickerType.INDICES:
                tissue_picker = TissuePickerIndices(
                    cells_index=tissue_profile.picker_cells_index)
            elif tissue_profile.picker_type is CellsPickerType.PERCENT:
                tissue_picker = TissuePickerPercent(
                    cells_percent=tissue_profile.picker_cells_percent)
            else:
                raise BetseSimTissueException(
                    'Tissue profile picker type "{}" unrecognized.'.format(
                        tissue_profile.picker_type))

            # Map this profile's name to a high-level tissue profile object.
            self.tissue_name_to_profile[tissue_profile.name] = TissueProfile(
                name=tissue_profile.name,
                z_order=tissue_z_order,
                picker=tissue_picker,
                is_gj_insular=tissue_profile.is_gj_insular,
                Dm_Na=tissue_profile.Dm_Na,
                Dm_K=tissue_profile.Dm_K,
                Dm_Cl=tissue_profile.Dm_Cl,
                Dm_Ca=tissue_profile.Dm_Ca,
                Dm_M=tissue_profile.Dm_M,
                Dm_P=tissue_profile.Dm_P,
            )


    @type_check
    def _init_cuts(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Internally initialize all cut profiles defined by the passed simulation
        configuration.

        Specifically, this method encapsulates each low-level YAML-formatted
        list item defining a cut profile in this configuration with a
        high-level instance of the :class:`CellsProfileABC` class.

        Parameters
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        # Ordered dictionaries mapping from the name of each each cut profile
        # to the corresponding profile.
        self.cut_name_to_profile = OrderedDict()

        # If cut profiles are enabled...
        if p.is_tissue_profiles:
            # For each low-level cut profile...
            for cut_index, cut_profile in enumerate(p.cut_profiles):
                # If this profile is non-unique, raise an exception.
                if cut_profile.name in self.cut_name_to_profile:
                    raise BetseSimTissueException(
                        'Cut profile "{0}" non-unique '
                        '(i.e., two or more cut profiles named "{0}").'.format(
                            cut_profile.name))

                # Map this profile's name to a high-level profile object.
                self.cut_name_to_profile[cut_profile.name] = CutProfile(
                    name=cut_profile.name,
                    z_order=cut_index + 1,
                    picker=TissuePickerImage(
                        filename=cut_profile.picker_image_filename,
                        dirname=p.conf_dirname)
                )

        # If the cutting event is enabled, initialize this event *AFTER*
        # initializing cut profiles, as the former assumes the latter to exist.
        if bool(p._conf['cutting event']['event happens']):
            # If one or more cut profiles are also enabled...
            if self.cut_name_to_profile:
                # For the name of each such profile applied by this event...
                for cut_profile_name in p.event_cut_profile_names:
                    # If this cut profile does *NOT* exist, raise an exception.
                    if cut_profile_name not in self.cut_name_to_profile:
                        raise BetseSimTissueException(
                            'Cut profile "{}" referenced by '
                            'cutting event not found.'.format(
                                cut_profile_name))

                # Define this event.
                self.event_cut = SimEventCut(p=p)
            # Else, no cut profiles are also enabled. Since this effectively
            # disables this event, warn the user that this is the case.
            else:
                logs.log_warning(
                    'Ignoring cutting event, as cut profiles are disabled.')


    @type_check
    def _init_events(self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Initialize all scheduled interventions for the passed simulation
        configuration.

        Parameters
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        self._init_events_global(p)


    @type_check
    def _init_events_global(
        self, p: 'betse.science.parameters.Parameters') -> None:
        '''
        Initialize all **global scheduled interventions** (i.e., events
        globally applicable to all cells) for the passed simulation
        configuration.

        Parameters
        ----------
        p : betse.science.parameters.Parameters
            Current simulation configuration.
        '''

        # If the voltage event is enabled...
        if bool(p._conf['apply external voltage']['event happens']):
            # If extracellular spaces are also enabled, enable this event.
            # if p.is_ecm:
            self._event_voltage = SimEventPulseVoltage(p=p)
            # Else, log a non-fatal warning.
            # else:
            #     logs.log_warning(
            #         'Ignoring voltage event, '
            #         'as extracellular spaces are disabled.')

    #FIXME: Refactor as follows:
    #
    #* Shift the body of the existing init_events() method into this method.
    #* Replace all calls to the init_events() method with calls to this method.
    #
    #Previously, we considered renaming this method to
    #_map_tissue_profiles_to_cells(). Doing so appears wholly insufficient,
    #however; external callers *MUST* finalize the initialization of this
    #object after the cell cluster has been created. See the
    #SimPhase.__init__() method for further (albeit obsolete) discussion.
    @type_check
    def init_profiles(self, phase: SimPhase) -> None:
        '''
        Finalize the initialization of all events and tissue and cut profiles
        for the passed simulation phase.

        Specifically, this method:

        * Defines all dictionaries mapping from tissue profile names to
          one-dimensional arrays of the indices of various cellular objects
          (e.g., cells, cell membranes) in the cluster belonging to these
          tissues. This includes the :attr:`cell_target_inds`,
          :attr:`env_target_inds`, and :attr:`tissue_target_inds` dictionaries.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        # Log this initialization.
        logs.log_debug('Mapping tissue, surgery, and boundary profiles...')

        # Localize pertinent simulation phase objects for convenience.
        cells = phase.cells
        p     = phase.p
        sim   = phase.sim

        #FIXME: These three dictionaries are all indexed by tissue profile names
        #and hence should simply be folded into the "TissueProfile" class as
        #instance variables whose values are the corresponding one-dimensional
        #Numpy lists for that specific tissue. This eliminates awkward lookups.
        self.cell_target_inds = {}
        self.env_target_inds = {}

        #FIXME: Let's rename this to something less ambiguous. Since it indexes
        #cell membranes, "mem_target_inds" might be a reasonable name? Lo-ho-ho!
        self.tissue_target_inds = {}

        # For each tissue profiles defined by this simulation configuration...
        for tissue_name, tissue_profile in self.tissue_name_to_profile.items():
            # One-dimensional Numpy arrays of the indices of all cells and cell
            # membranes comprising this tissue.
            tissue_cells_index, tissue_mems_index = (
                tissue_profile.picker.pick_cells_and_mems(cells=cells, p=p))

            # Persist these arrays for this tissue.
            self.cell_target_inds[tissue_name] = tissue_cells_index
            self.tissue_target_inds[tissue_name] = tissue_mems_index

            # If this tissue is non-empty (i.e., contains at least one cell)...
            if len(self.cell_target_inds[tissue_name]):
                # Get ECM targets.
                if p.is_ecm:
                    ecm_targs_mem = list(cells.map_mem2ecm[tissue_mems_index])
                    self.env_target_inds[tissue_name] = ecm_targs_mem

                # Set the values of Dmems and ECM diffusion based on the
                # identified target indices.
                if p.ions_dict['Na'] == 1:
                    sim.Dm_cells[sim.iNa][tissue_mems_index] = (
                        tissue_profile.Dm_Na)

                if p.ions_dict['K'] == 1:
                    sim.Dm_cells[sim.iK][tissue_mems_index] = (
                        tissue_profile.Dm_K)

                if p.ions_dict['Cl'] == 1:
                    sim.Dm_cells[sim.iCl][tissue_mems_index] = (
                        tissue_profile.Dm_Cl)

                if p.ions_dict['Ca'] == 1:
                    sim.Dm_cells[sim.iCa][tissue_mems_index] = (
                        tissue_profile.Dm_Ca)

                if p.ions_dict['M'] == 1:
                    sim.Dm_cells[sim.iM][tissue_mems_index] = (
                        tissue_profile.Dm_M)

                if p.ions_dict['P'] == 1:
                    sim.Dm_cells[sim.iP][tissue_mems_index] = (
                        tissue_profile.Dm_P)

    # ..................{ INITIALIZERS ~ event              }..................
    #FIXME: The overwhelming majority of the logic performed by this method
    #should be shifted into the _init_events() method. The only logic that
    #should remain here is that which requires a fully seeded cell cluster --
    #notably, the logic initializing the "gj_block" event in the
    #_init_events_global() method. All other logic is simple and unconditional.
    #FIXME: Shift what remains of this method into the init_profiles() method.
    #See that method's FIXME for further details.
    @type_check
    def init_events(self, phase: SimPhase) -> None:
        '''
        Initialize all scheduled interventions for the passed simulation phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        self._finalize_events_global(phase)
        self._finalize_events_tissue(phase)


    @type_check
    def _finalize_events_global(self, phase: SimPhase) -> None:
        '''
        Initialize all **global scheduled interventions** (i.e., events
        globally applicable to all cells) for the passed simulation phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        # Localize pertinent simulation phase objects for convenience.
        cells = phase.cells
        p     = phase.p

        if p.global_options['K_env'] != 0:
            self.t_on_Kenv = p.global_options['K_env'][0]
            self.t_off_Kenv = p.global_options['K_env'][1]
            self.t_change_Kenv = p.global_options['K_env'][2]
            self.mem_mult_Kenv = p.global_options['K_env'][3]

        if p.global_options['Cl_env'] != 0:
            self.t_on_Clenv = p.global_options['Cl_env'][0]
            self.t_off_Clenv = p.global_options['Cl_env'][1]
            self.t_change_Clenv = p.global_options['Cl_env'][2]
            self.mem_mult_Clenv = p.global_options['Cl_env'][3]

        if p.global_options['Na_env'] != 0:
            self.t_on_Naenv = p.global_options['Na_env'][0]
            self.t_off_Naenv = p.global_options['Na_env'][1]
            self.t_change_Naenv = p.global_options['Na_env'][2]
            self.mem_mult_Naenv = p.global_options['Na_env'][3]

        if p.global_options['T_change'] != 0:
            self.tonT = p.global_options['T_change'][0]
            self.toffT = p.global_options['T_change'][1]
            self.trampT = p.global_options['T_change'][2]
            self.multT = p.global_options['T_change'][3]

        if p.global_options['gj_block'] != 0:
            self.tonGJ = p.global_options['gj_block'][0]
            self.toffGJ = p.global_options['gj_block'][1]
            self.trampGJ = p.global_options['gj_block'][2]

            numo = int(p.global_options['gj_block'][3])

            if numo > 100:
                numo = 100
            elif numo < 1:
                numo = 1

            #FIXME: Isn't the exact same "data_length" already available as the
            #"self.data_length" variable initialized by the __init__() method?
            #Leaping unicorns deny the inevitability of love!
            data_length = len(cells.mem_i)
            data_fraction = int((numo/100)*data_length)

            mem_i_copy = np.copy(cells.mem_i[:])
            shuffle(mem_i_copy)
            self.targets_gj_block = [
                cells.mem_i[x] for x in range(0,data_fraction)]

        if p.global_options['NaKATP_block'] != 0:
            self.tonNK = p.global_options['NaKATP_block'][0]
            self.toffNK = p.global_options['NaKATP_block'][1]
            self.trampNK = p.global_options['NaKATP_block'][2]


    @type_check
    def _finalize_events_tissue(self, phase: SimPhase) -> None:
        '''
        Initialize all **targeted scheduled interventions** (i.e., events only
        applicable to specific tissue profiles) for the passed simulation
        phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        # Localize pertinent simulation phase objects for convenience.
        cells = phase.cells
        p = phase.p

        if p.scheduled_options['Na_mem'] != 0:
            self.t_on_Namem = p.scheduled_options['Na_mem'][0]
            self.t_off_Namem = p.scheduled_options['Na_mem'][1]
            self.t_change_Namem = p.scheduled_options['Na_mem'][2]
            self.mem_mult_Namem = p.scheduled_options['Na_mem'][3]
            self.apply_Namem = p.scheduled_options['Na_mem'][4]
            self.function_Namem = p.scheduled_options['Na_mem'][5]

            self.targets_Namem = []
            for profile in self.apply_Namem:
                targets = self.tissue_target_inds[profile]
                self.targets_Namem.append(targets)

            self.targets_Namem = [
                item for sublist in self.targets_Namem for item in sublist]

            self.scalar_Namem = 1
            self.dyna_Namem = lambda t: 1

            # call a special toolbox function to change membrane permeability: spatial grads
            # 'gradient_x', 'gradient_y', 'gradient_r'

            if self.function_Namem != 'None':
                self.scalar_Namem, self.dyna_Namem = getattr(
                    mod, self.function_Namem)(self.targets_Namem,cells,p)

        #----------------------------------------------

        if p.scheduled_options['K_mem'] != 0:
            self.t_on_Kmem = p.scheduled_options['K_mem'][0]
            self.t_off_Kmem = p.scheduled_options['K_mem'][1]
            self.t_change_Kmem = p.scheduled_options['K_mem'][2]
            self.mem_mult_Kmem = p.scheduled_options['K_mem'][3]
            self.apply_Kmem = p.scheduled_options['K_mem'][4]
            self.function_Kmem = p.scheduled_options['K_mem'][5]

            self.targets_Kmem = []
            for profile in self.apply_Kmem:
                targets = self.tissue_target_inds[profile]
                self.targets_Kmem.append(targets)

            self.targets_Kmem = [
                item for sublist in self.targets_Kmem for item in sublist]

            self.scalar_Kmem = 1
            self.dyna_Kmem = lambda t: 1

            if self.function_Kmem != 'None':
                # call a special toolbox function to change membrane permeability: spatial grads
                # 'gradient_x', 'gradient_y', 'gradient_r'

                self.scalar_Kmem, self.dyna_Kmem = getattr(
                    mod, self.function_Kmem)(self.targets_Kmem, cells, p)

        #----------------------------------------------

        if p.scheduled_options['Cl_mem'] != 0:
            self.t_on_Clmem = p.scheduled_options['Cl_mem'][0]
            self.t_off_Clmem = p.scheduled_options['Cl_mem'][1]
            self.t_change_Clmem = p.scheduled_options['Cl_mem'][2]
            self.mem_mult_Clmem = p.scheduled_options['Cl_mem'][3]
            self.apply_Clmem = p.scheduled_options['Cl_mem'][4]
            self.function_Clmem = p.scheduled_options['Cl_mem'][5]

            self.targets_Clmem = []
            for profile in self.apply_Clmem:
                targets = self.tissue_target_inds[profile]
                self.targets_Clmem.append(targets)

            self.targets_Clmem = [
                item for sublist in self.targets_Clmem for item in sublist]

            self.scalar_Clmem = 1
            self.dyna_Clmem = lambda t: 1

            # Call a special toolbox function to change membrane permeability:
            # spatial grads ('gradient_x', 'gradient_y', 'gradient_r').
            if self.function_Clmem != 'None':
                self.scalar_Clmem, self.dyna_Clmem = getattr(
                    mod, self.function_Clmem)(self.targets_Clmem, cells, p)

        #----------------------------------------------

        if p.scheduled_options['Ca_mem'] != 0:
            self.t_on_Camem = p.scheduled_options['Ca_mem'][0]
            self.t_off_Camem = p.scheduled_options['Ca_mem'][1]
            self.t_change_Camem = p.scheduled_options['Ca_mem'][2]
            self.mem_mult_Camem = p.scheduled_options['Ca_mem'][3]
            self.apply_Camem = p.scheduled_options['Ca_mem'][4]
            self.function_Camem = p.scheduled_options['Ca_mem'][5]

            self.targets_Camem = []
            for profile in self.apply_Camem:
                targets = self.tissue_target_inds[profile]
                self.targets_Camem.append(targets)

            self.targets_Camem = [
                item for sublist in self.targets_Camem for item in sublist]

            self.scalar_Camem = 1
            self.dyna_Camem = lambda t: 1

            if self.function_Camem != 'None':
                # call a special toolbox function to change membrane permeability: spatial grads
                # 'gradient_x', 'gradient_y', 'gradient_r'
                self.scalar_Camem, self.dyna_Camem = getattr(
                    mod, self.function_Camem)(self.targets_Camem, cells, p)

        #----------------------------------------------

        if p.scheduled_options['pressure'] != 0:

            self.t_onP = p.scheduled_options['pressure'][0]
            self.t_offP = p.scheduled_options['pressure'][1]
            self.t_changeP = p.scheduled_options['pressure'][2]
            self.rate_P = p.scheduled_options['pressure'][3]
            self.apply_P = p.scheduled_options['pressure'][4]
            self.function_P = p.scheduled_options['pressure'][5]

            self.targets_P = []
            for profile in self.apply_P:
                targets = self.cell_target_inds[profile]
                self.targets_P.append(targets)

            self.targets_P = [
                item for sublist in self.targets_P for item in sublist]

            self.scalar_P = 1
            self.dyna_P = lambda t: 1

            # call a special toolbox function to change membrane permeability: spatial grads
            # 'gradient_x', 'gradient_y', 'gradient_r'
            if self.function_P != 'None':
                self.scalar_P, self.dyna_P = getattr(mod, self.function_P)(
                    self.targets_P, cells, p)

        #--------------------------------------------------------

        if p.scheduled_options['ecmJ'] != 0 and p.is_ecm:
            self.t_on_ecmJ  = p.scheduled_options['ecmJ'][0]
            self.t_off_ecmJ = p.scheduled_options['ecmJ'][1]
            self.t_change_ecmJ = p.scheduled_options['ecmJ'][2]
            self.apply_ecmJ = p.scheduled_options['ecmJ'][3]
            self.mult_ecmJ = p.scheduled_options['ecmJ'][4]

            self.targets_ecmJ = []
            for profile in self.apply_ecmJ:
                targets = self.env_target_inds[profile]
                self.targets_ecmJ.append(targets)

            self.targets_ecmJ = [
                item for sublist in self.targets_ecmJ for item in sublist]

    # ..................{ FIRERS                            }..................
    @type_check
    def fire_events(self, phase: SimPhase, t: float) -> None:
        '''
        Apply all tissue manipulations for the passed time step of the passed
        simulation phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        t : float
            Current time step of this phase being simulated.
        '''

        self._fire_events_global(phase=phase, t=t)
        self._fire_events_tissue(phase=phase, t=t)
        self.makeAllChanges(phase.sim)


    @type_check
    def _fire_events_global(self, phase: SimPhase, t: float) -> None:
        '''
        Apply all **global scheduled interventions** (i.e., events globally
        applicable to all cells) for the passed time step of the passed
        simulation phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        t : float
            Current time step of this phase being simulated.
        '''

        # Localize pertinent simulation phase objects for convenience.
        p   = phase.p
        sim = phase.sim

        if p.global_options['K_env'] != 0:
            effector_Kenv = tb.pulse(
                t, self.t_on_Kenv, self.t_off_Kenv, self.t_change_Kenv)

            if p.is_ecm: # simulate addition of potassium salt to remain charge neutral
                sim.c_env_bound[sim.iK] = (
                    self.mem_mult_Kenv*effector_Kenv*p.env_concs['K'] +
                    p.env_concs['K']
                )
                sim.c_env_bound[sim.iM] = (
                    self.mem_mult_Kenv*effector_Kenv*p.env_concs['K'] +
                    p.env_concs['M']
                )
            else:
                sim.cc_env[sim.iK][:] = (
                    self.mem_mult_Kenv*effector_Kenv*p.conc_env_k + p.conc_env_k)

        if p.global_options['Cl_env'] != 0 and p.ions_dict['Cl'] == 1:
            effector_Clenv = tb.pulse(t,self.t_on_Clenv,self.t_off_Clenv,self.t_change_Clenv)

            if not p.is_ecm:
                sim.cc_env[sim.iCl][:] = self.mem_mult_Clenv*effector_Clenv*p.conc_env_cl + p.conc_env_cl

            else:  # simulate addition of sodium chloride to remain charge neutral
                sim.c_env_bound[sim.iCl] = self.mem_mult_Clenv*effector_Clenv*p.env_concs['Cl'] + p.env_concs['Cl']
                sim.c_env_bound[sim.iNa] = self.mem_mult_Clenv*effector_Clenv*p.env_concs['Cl'] + p.env_concs['Na']

        if p.global_options['Na_env'] != 0:
            effector_Naenv = tb.pulse(t,self.t_on_Naenv,self.t_off_Naenv,self.t_change_Naenv)

            if not p.is_ecm:
                sim.cc_env[sim.iNa][:] = self.mem_mult_Naenv*effector_Naenv*p.conc_env_na + p.conc_env_na

            else: # simulate addition of sodium salt to remain charge neutral
                sim.c_env_bound[sim.iNa] = self.mem_mult_Naenv*effector_Naenv*p.env_concs['Na'] + p.env_concs['Na']
                sim.c_env_bound[sim.iM] = self.mem_mult_Naenv*effector_Naenv*p.env_concs['Na'] + p.env_concs['M']

        if p.global_options['T_change'] != 0:
            sim.T = self.multT*tb.pulse(t,self.tonT,self.toffT,self.trampT)*p.T + p.T

        if p.global_options['gj_block'] != 0:
            sim.gj_block[self.targets_gj_block] = (1.0 - tb.pulse(t,self.tonGJ,self.toffGJ,self.trampGJ))

        if p.global_options['NaKATP_block'] != 0:
            sim.NaKATP_block = (1.0 - tb.pulse(t,self.tonNK,self.toffNK,self.trampNK))

    # ..................{ FIRERS ~ tissue                   }..................
    @type_check
    def _fire_events_tissue(self, phase: SimPhase, t: float) -> None:
        '''
        Apply all **targeted scheduled interventions** (i.e., events only
        applicable to specific tissue profiles) for the passed time step of the
        passed simulation phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        t : float
            Current time step of this phase being simulated.
        '''

        # Localize pertinent simulation phase objects for convenience.
        cells = phase.cells
        p = phase.p
        sim = phase.sim

        if p.scheduled_options['Na_mem'] != 0:
            effector_Na = (
                self.scalar_Namem*self.dyna_Namem(t)*
                tb.pulse(t,self.t_on_Namem,self.t_off_Namem,self.t_change_Namem))
            sim.Dm_scheduled[sim.iNa][self.targets_Namem] = (
                self.mem_mult_Namem*effector_Na*self.tissue_default.Dm_Na)

        if p.scheduled_options['K_mem'] != 0:
            effector_K = (
                self.scalar_Kmem*self.dyna_Kmem(t)*
                tb.pulse(t,self.t_on_Kmem,self.t_off_Kmem,self.t_change_Kmem))
            sim.Dm_scheduled[sim.iK][self.targets_Kmem] = (
                self.mem_mult_Kmem*effector_K*self.tissue_default.Dm_K)

        if p.scheduled_options['Cl_mem'] != 0 and p.ions_dict['Cl'] != 0:
            effector_Cl = (
                self.scalar_Clmem*self.dyna_Clmem(t)*
                tb.pulse(t,self.t_on_Clmem,self.t_off_Clmem,self.t_change_Clmem))
            sim.Dm_scheduled[sim.iCl][self.targets_Clmem] = (
                self.mem_mult_Clmem*effector_Cl*self.tissue_default.Dm_Cl)

        if p.scheduled_options['Ca_mem'] != 0 and p.ions_dict['Ca'] != 0:
            effector_Ca = (
                self.scalar_Camem*self.dyna_Camem(t)*
                tb.pulse(t,self.t_on_Camem,self.t_off_Camem,self.t_change_Camem))
            sim.Dm_scheduled[sim.iCa][self.targets_Camem] = (
                self.mem_mult_Camem*effector_Ca*self.tissue_default.Dm_Ca)

        if p.scheduled_options['pressure'] != 0:
            # logs.log_debug('Applying pressure event...')
            sim.P_mod[self.targets_P] = (
                self.scalar_P*self.dyna_P(t)*self.rate_P*tb.pulse(
                    t,self.t_onP, self.t_offP, self.t_changeP))

        if p.scheduled_options['ecmJ'] != 0:
            if p.is_ecm:
                for i, _ in enumerate(sim.D_env):
                    effector_ecmJ = self.mult_ecmJ*tb.pulse(
                        t,self.t_on_ecmJ,self.t_off_ecmJ,self.t_change_ecmJ)

                    sim.D_env[i][self.targets_ecmJ] = (
                        sim.D_env_base[i][self.targets_ecmJ]*(
                            1 - effector_ecmJ) + effector_ecmJ*sim.D_free[i])

                    sim.D_env_weight = sim.D_env_weight.ravel()
                    sim.D_env_weight_base = sim.D_env_weight_base.ravel()

                    sim.D_env_weight[self.targets_ecmJ] = (
                        sim.D_env_weight_base[
                            self.targets_ecmJ]*(1-effector_ecmJ) +
                        effector_ecmJ)

                    sim.D_env_weight = sim.D_env_weight.reshape(cells.X.shape)
                    sim.D_env_weight_base = sim.D_env_weight_base.reshape(cells.X.shape)

            else:
                effector_ecmJ = self.mult_ecmJ * tb.pulse(
                    t, self.t_on_ecmJ, self.t_off_ecmJ, self.t_change_ecmJ)

                sim.D_env_weight = sim.D_env_weight.ravel()
                sim.D_env_weight_base = sim.D_env_weight_base.ravel()

                sim.D_env_weight[self.targets_ecmJ] = (
                    sim.D_env_weight_base[
                        self.targets_ecmJ]*(1-effector_ecmJ) +
                    effector_ecmJ)

                sim.D_env_weight = sim.D_env_weight.reshape(cells.X.shape)
                sim.D_env_weight_base = sim.D_env_weight_base.reshape(cells.X.shape)

        # If...
        if (
            # The cutting event is enabled...
            self.event_cut is not None and
            # The cutting event has yet to be performed...
            not self.event_cut.is_fired and
            # The current time step is at least the time step at which the
            # cutting event is scheduled to occur...
            t >= p.event_cut_time
        ):
            for cut_profile_name in p.event_cut_profile_names:
                logs.log_info(
                    'Cutting cell cluster via cut profile "%s"...',
                    cut_profile_name)

                cut_profile_picker = (
                    self.cut_name_to_profile[cut_profile_name].picker)
                self._cut_cells(phase=phase, tissue_picker=cut_profile_picker)

            logs.log_info('Cutting event successful! Resuming simulation...')

            # Redo main data length variable for this dynamics module with
            # updated world.
            self.data_length = len(cells.mem_i)

            self.init_profiles(phase)
            cells.redo_gj(phase)
            self.init_events(phase)

            # Avoid repeating this cutting event at subsequent time steps.
            self.event_cut.fire(phase=phase, time_step=t)

        # If the voltage event is enabled, adjust the voltage accordingly.
        if self._event_voltage is not None:
            self._event_voltage.fire(phase=phase, time_step=t)


    @type_check
    def _cut_cells(
        self,
        phase: SimPhase,
        tissue_picker: TissuePickerABC,
    ) -> None:
        '''
        Permanently remove all cells selected by the passed tissue picker from
        the cell cluster described by the passed simulation phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        tissue_picker : TissuePickerABC
            Object matching all cells to be removed.
        '''

        # Localize pertinent simulation phase objects for convenience.
        cells = phase.cells
        p = phase.p
        sim = phase.sim

        # Redo environmental diffusion matrices by setting the environmental spaces
        # around cut world to the free value (True) or not (False)?
        # open_TJ = True

        # Subtract this bitmap's clipping mask from the global cluster mask.
        # bitmap_mask = tissue_picker.get_image_mask(cells).clipping_matrix
        # cells.cluster_mask = cells.cluster_mask - bitmap_mask

        # FIXME, if deformation is too much, the following line will crash as
        # the "target_inds_cell" is null. Rayse an uman weedable eggseption.

        # Indices of all cells and cell membranes to be removed.
        target_inds_cell, target_inds_mem = tissue_picker.pick_cells_and_mems(
            cells=cells, p=p)
            # dmem_list = tissue_profile['diffusion constants']

        # get the corresponding flags to membrane entities
        target_inds_gj,_,_ = tb.flatten(cells.cell_to_nn_full[target_inds_cell])

        if p.is_ecm:
            # get environmental targets around each removed cell:
            ecm_targs_cell = list(cells.map_cell2ecm[target_inds_cell])
            ecm_targs_mem = list(cells.map_mem2ecm[target_inds_mem])

            ecm_targs = []
            for v in ecm_targs_cell:
                ecm_targs.append(v)
            for v in ecm_targs_mem:
                ecm_targs.append(v)

            # if sim.molecules is not None and 'ATP' in sim.molecules.core.molecules:
            #
            #     # get concentration of ATP in cells to be removed:
            #     cell_ATP = sim.molecules.core.cell_concs['ATP'][target_inds_cell] * cells.cell_vol[target_inds_cell]
            #     # move this entire concentration to the extracellular spaces (assumed upon cell bursting)
            #     sim.molecules.core.env_concs['ATP'][ecm_targs_cell] = cell_ATP / (p.cell_height * cells.delta ** 2)
            #
            # elif sim.grn is not None and 'ATP' in sim.grn.core.molecules:
            #
            #     # get concentration of ATP in cells to be removed:
            #     cell_ATP = sim.grn.core.cell_concs['ATP'][target_inds_cell] * cells.cell_vol[target_inds_cell]
            #     # move this entire concentration to the extracellular spaces (assumed upon cell bursting)
            #     sim.grn.core.env_concs['ATP'][ecm_targs_cell] = cell_ATP / (p.cell_height * cells.delta ** 2)

            # redo environmental diffusion matrices by
            # setting the environmental spaces around cut world to the free value -- if desired!:
            # if open_TJ is True:
        # save the x,y coordinates of the original boundary cell and membrane points:
        # old_bflag_cellxy = np.copy(cells.cell_centres[cells.bflags_cells])
        # old_bflag_memxy = np.copy(cells.mem_mids_flat[cells.bflags_mems])

        # set up the situation to make world joined to cut world have more permeable membranes:
        hurt_cells = np.zeros(len(cells.cell_i))

        # If we're creating a wound-induced channel ------------------------------------------------
        target_inds_gj_unique = np.unique(target_inds_gj)

        for i, inds in enumerate(cells.cell_to_nn_full): # for all the nn inds to a cell...
            inds_array = np.asarray(inds)
            inds_in_target = np.intersect1d(inds_array,target_inds_gj_unique)

            if len(inds_in_target):
                hurt_cells[i] = 1  # flag the cell as a "hurt" cell

        hurt_inds = (hurt_cells == 1).nonzero()
        sim.hurt_mask = np.zeros(sim.cdl)
        sim.hurt_mask[hurt_inds] = 1.0

        #----------------------------------------------------------------------------------

        # Names of all attributes in the current "Simulation" object.
        sim_names = list(sim.__dict__.keys())

        #FIXME: Consider redeclaring this list as a set instead, renaming that
        #set to "specials_names", and removing the duplicate declaration of
        #that variable below. Hot jumpin' jallopies!

        # Names of all such attributes to be repaired due to this cutting event.
        specials_list = [
            'cc_cells',
            'cc_mems',
            'cc_env',
            'cc_er',
            'z_array',
            'z_array_er',
            'fluxes_gj',
            'fluxes_mem',
            'Dm_base',
            'Dm_cells',
            'Dm_scheduled',
            'Dm_vg',
            'Dm_cag',
            'Dm_morpho',
            'Dm_er_base',
            'Dm_er_CICR',
            'Dm_stretch',
            'Dm_custom',
            'D_gj',
            'fluxes_intra',
            'cc_at_mem',
        ]

        if p.is_ecm:
            specials_list.remove('cc_env')
            extra = ['z_array_cells']
            for ent in extra:
                specials_list.append(ent)

        special_names = set(specials_list)

        #FIXME: This is awesome-sauce in a BETSE jar, but I don't quite grok what's
        #going on. It'd be swell if this could recieve some documentation
        #massaging. The tidal waves of time recede, inch by minute inch!
        for name in sim_names:
            if name in special_names: # if this is a nested data structure...
                super_data = getattr(sim, name)
                super_data2 = []

                for i, data in enumerate(super_data):
                    if isinstance(data, np.ndarray):
                        if len(data) == len(cells.cell_i):
                            data2 = np.delete(data,target_inds_cell)

                        elif len(data) == len(cells.mem_i):
                            data2 = np.delete(data,target_inds_mem)

                        # elif len(data) == len(cells.nn_i):
                        #     data2 = np.delete(data,target_inds_gj)

                    elif isinstance(data, list):
                        data2 = []
                        if len(data) == len(cells.cell_i):
                            for index in sorted(target_inds_cell, reverse=True):
                                del data[index]
                            data2.append(data[index])

                        elif len(data) == len(cells.mem_i):
                            for index in sorted(target_inds_mem, reverse=True):
                                del data[index]
                            data2.append(data[index])

                    super_data2.append(data2)

                if type(super_data) == np.ndarray:
                    super_data2 = np.asarray(super_data2)

                setattr(sim, name, super_data2)

            else:
                data = getattr(sim, name)

                if isinstance(data, np.ndarray):
                    if len(data) == len(cells.cell_i):
                        data2 = np.delete(data, target_inds_cell)
                        setattr(sim, name, data2)

                    elif len(data) == len(cells.mem_i):
                        data2 = np.delete(data, target_inds_mem)
                        setattr(sim, name, data2)

                elif isinstance(data, list):
                    index = None
                    data2 = []

                    if len(data) == len(cells.cell_i):
                        for index in sorted(target_inds_cell, reverse=True):
                            del data[index]
                        data2.append(data[index])
                        setattr(sim, name, data2)

                    elif len(data) == len(cells.mem_i):
                        for index in sorted(target_inds_mem, reverse=True):
                            del data[index]
                        data2.append(data[index])
                        setattr(sim, name, data2)

        if p.Ca_dyn and sim.endo_retic is not None:
            sim.endo_retic.remove_ers(sim, target_inds_cell)

        #------------------------------ Fix-up cell world ---------------------
        new_cell_centres = []
        new_ecm_verts = []
        removal_flags = np.zeros(len(cells.cell_i))
        removal_flags[target_inds_cell] = 1

        for i, flag in enumerate(removal_flags):
            if flag == 0:
                new_cell_centres.append(cells.cell_centres[i])
                new_ecm_verts.append(cells.ecm_verts[i])

        cells.cell_centres = np.asarray(new_cell_centres)
        cells.ecm_verts = np.asarray(new_ecm_verts, dtype=object)

        # recalculate ecm_verts_unique:
        ecm_verts_flat,_,_ = tb.flatten(cells.ecm_verts)
        ecm_verts_set = set()

        for vert in ecm_verts_flat:
            ptx = vert[0]
            pty = vert[1]
            ecm_verts_set.add((ptx,pty))

        cells.ecm_verts_unique = [list(verts) for verts in list(ecm_verts_set)]
        cells.ecm_verts_unique = np.asarray(cells.ecm_verts_unique)  # convert to numpy array

        #-----------------------------------------------------------------
        logs.log_info('Recalculating cluster variables for new configuration...')

        # Cut the DEC meshes and save the inds to modify data defined on DEC
        # edges, verts, and simplices.
        self.targets_dec_edges, self.targets_dec_tcells = cells.mesh.cut_mesh(
            target_inds_cell)
        self.targets_dec_tverts = target_inds_cell

        cells.cellVerts(p)   # create individual cell polygon vertices and other essential data structures
        cells.cellMatrices(p)  # creates a variety of matrices used in routine cells calculations
        cells.intra_updater(p)  # creates matrix used for finite volume integration on cell patch
        cells.cell_vols(p)  # calculate the volume of cell and its internal regions
        cells.mem_processing(p)  # calculates membrane nearest neighbours, ecm interaction, boundary tags, etc
        cells.near_neigh(p)  # Calculate the nn array for each cell
        cells.calc_gj_vects(p)
        cells.environment(p)  # define features of the ecm grid
        cells.make_maskM(p)
        cells.grid_len = len(cells.xypts)

        # update sim data lengths with new world values:
        sim.mdl = len(cells.mem_mids_flat)  # mems-data-length
        sim.cdl = len(cells.cell_centres)  # cells-data-length

        if p.is_ecm:  # set environnment data length
            sim.edl = len(cells.xypts)
        else:
            sim.edl = len(cells.mem_mids_flat)

        logs.log_info('Re-creating cell network Poisson solver...')
        cells.graphLaplacian(p)

        # # if microtubules
        # if sim.mtubes is not None:
        #     sim.mtubes.remove_mtubes(
        #         target_inds_mem, target_inds_cell, cells, sim, p)

        # if running voltage gated gap junctions, reinnitialize them:
        if p.v_sensitive_gj and sim.gj_funk is not None:
            sim.gj_funk.init(sim, cells, p)

        # delete data from molecules objects:
        if p.molecules_enabled and sim.molecules is not None:
            sim.molecules.core.mod_after_cut_event(
                phase, target_inds_cell, target_inds_mem)

        if p.grn_enabled and sim.grn is not None:
            sim.grn.core.mod_after_cut_event(
                phase, target_inds_cell, target_inds_mem)

        # Save target inds so they can be used outside of the core simulator
        # (i.e. dynamically in "sim-grn").
        sim.target_inds_cell_o = target_inds_cell
        sim.target_inds_mem_o = target_inds_mem

        # if hole_tag is False: # if we're not defining a hole at the beginning, reassign to new bflags
        sim.initDenv(cells,p)

        sim.conc_J_x = np.zeros(len(cells.xypts))
        sim.conc_J_y = np.zeros(len(cells.xypts))

        if p.fluid_flow or p.deformation:
            # make a laplacian and solver for discrete transfers on closed, irregular cell network:
            cells.deform_tools(p)

            if p.deformation:  # if user desires deformation:

                # create a copy of cells world, to apply deformations to for visualization purposes only:
                sim.cellso = copy.deepcopy(cells)

                if p.td_deform:
                    # make a laplacian and solver for discrete transfers on closed, irregular cell network
                    logs.log_info('Creating cell network Poisson solver...')
                    cells.graphLaplacian(p)

        # if p.sim_eosmosis is True:
        #
        #     if sim.move_pumps_channels is not None:
        #
        #         sim.move_pumps_channels.remove_data(target_inds_cell)

        # Calculate targets for wound channel.
        match_inds = (sim.hurt_mask == 1.0).nonzero()
        mem_match_inds = tb.flatten(cells.cell_to_mems[match_inds[0]])[0]
        self.targets_vgWound = mem_match_inds

        if p.break_TJ: # we need to redo the TJ barrier at the wound site:
            # first identify indices of exterior and interior extracellular grid sites corresponding to the wound TJ:
            cell_wound_inds = (sim.hurt_mask == 1.0).nonzero()
            neigh_to_wound_cells, _, _ = tb.flatten(cells.cell_nn[cell_wound_inds])

            interior_wound_inds = cells.cell_to_mems[neigh_to_wound_cells]
            interior_wound_inds, _, _ = tb.flatten(interior_wound_inds)

            ecmtarg_a = cells.map_mem2ecm[self.targets_vgWound]
            ecmtarg_b = cells.map_mem2ecm[interior_wound_inds]

            ecmtarg = np.hstack((ecmtarg_a, ecmtarg_b))
            # ecmtarg = ecmtarg_a

            Dw = np.copy(sim.D_env_weight.ravel())

            # assign the wound TJ to the user's requested value:
            Dw[ecmtarg] = 1.0*p.wound_TJ

            # re-assign the environmental weight matrix:
            sim.D_env_weight = Dw.reshape(cells.X.shape)*1

            # furthermore, recalculate the individual ion diffusion maps:
            if p.is_ecm:
                for i, dmat in enumerate(sim.D_env):
                    Denv_o = np.copy(dmat)
                    Denv_o[ecmtarg] = sim.D_free[i]*p.wound_TJ*sim.Dtj_rel[i]

                    # re-assign the ecm diffusion grid filled with the environmental values at wound site:
                    sim.D_env[i] = Denv_o*1.0

            # # re-calculate specific maps of conductivity in cells and environment
            # sim.sigma_cell = np.asarray([((z ** 2) * p.q * p.F * cc * D) / (p.kb * p.T) for (z, cc, D) in
            #                               zip(sim.zs, sim.cc_cells, sim.D_free)]).mean(axis=0)
            #
            # sim.sigma_env = np.asarray(
            #     [((z ** 2) * p.q * p.F * cc * D) / (p.kb * p.T) for (z, cc, D) in
            #      zip(sim.zs, sim.cc_env, sim.D_env)]).mean(
            #     axis=0)
            #
            # if p.is_ecm:
            #
            #     # environmental conductivity matrix needs to be smoothed to assured simulation stability:
            #     sim.sigma_env = gaussian_filter(sim.sigma_env.reshape(cells.X.shape), 1).ravel()

        # need to also re-do tissue profiles and GJ
        self.init_profiles(phase)
        cells.redo_gj(phase)  # redo gap junctions to isolate different tissue types

        # If this is the fast BETSE solver, re-initialize this solver.
        if p.solver_type is SolverType.FAST:
            sim.fast_sim_init(cells, p)

    # ..................{ CHANNELS                          }..................
    def stretchChannel(self,sim,cells,p,t):

        dd = np.sqrt(sim.d_cells_x**2 + sim.d_cells_y**2)

        # calculate strain from displacement
        eta = (dd/cells.R)

        # # create a smooth bivariate spline to interpolate deformation data from cells:
        # cellinterp_x = SmoothBivariateSpline(cells.cell_centres[:, 0], cells.cell_centres[:, 1], sim.d_cells_x, kx=1,
        #                                      ky=1)
        # cellinterp_y = SmoothBivariateSpline(cells.cell_centres[:, 0], cells.cell_centres[:, 1], sim.d_cells_y, kx=1,
        #                                      ky=1)
        #
        # # calculate deformations wrt the ecm using the smooth bivariate spline:
        # dmem_x = cellinterp_x.ev(cells.mem_mids_flat[:, 0], cells.mem_mids_flat[:, 1])
        # dmem_y = cellinterp_y.ev(cells.mem_mids_flat[:, 0], cells.mem_mids_flat[:, 1])
        #
        # # obtain normal component to membrane
        # dd = dmem_x*cells.mem_vects_flat[:,2] + dmem_y*cells.mem_vects_flat[:,3]
        #
        # # strain is the divergence of the displacement:
        # eta = np.dot(cells.M_sum_mems, dd*cells.mem_sa)/cells.cell_vol

        # self.active_NaStretch[self.targets_NaStretch] = tb.hill(sim.P_cells[cells.mem_to_cells][self.targets_NaStretch],
        #         self.NaStretch_halfmax,self.NaStretch_n)

        self.active_NaStretch[self.targets_NaStretch] = tb.hill(eta[cells.mem_to_cells][self.targets_NaStretch],
                self.NaStretch_halfmax,self.NaStretch_n)

        sim.Dm_stretch[sim.iNa] = self.maxDmNaStretch*self.active_NaStretch
        sim.Dm_stretch[sim.iK] = self.maxDmNaStretch*self.active_NaStretch


    def makeAllChanges(self, sim) -> None:
        '''
        Add together all effects to finalize changes to cell membrane
        permeabilities for the current time step.
        '''

        sim.Dm_cells = (
            sim.Dm_scheduled +
            sim.Dm_base
        )

        sim.P_cells = sim.P_mod + sim.P_base



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Global unit test configuration for BETSE.

:mod:`pytest` implicitly imports all functionality defined by this module into
all unit test modules. As this functionality includes all publicly declared
functional fixtures in this fixture subpackage, these tests may reference these
fixtures without explicit imports.
'''

# ....................{ IMPORTS ~ fixture : autouse       }....................
# Import fixtures automatically run at the start of the current test session,
# typically *NOT* manually required by specific tests, *AFTER* importing all
# non-autouse fixtures possibly required by these autouse fixtures above.

from betse_test._fixture.initter import betse_init_package



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **pathable** (i.e., commands reseding in the current ``${PATH}`` and
hence executable by specifying merely their basename rather than either
relative or absolute path) facilities.
'''

# ....................{ IMPORTS                           }....................
import shutil
from betse.exceptions import BetseCommandException
from betse.util.type.types import type_check, SequenceTypes, StrOrNoneTypes

# ....................{ EXCEPTIONS                        }....................
@type_check
def die_unless_pathable(
    command_basename: str, exception_message: StrOrNoneTypes = None):
    '''
    Raise an exception with the passed message unless an external command with
    the passed basename exists (i.e., unless an executable file with this
    basename resides in the current ``${PATH}``).

    Raises
    ----------
    BetseCommandException
        If no external command with this basename exists.
    '''

    # If this pathable is not found, raise an exception.
    if not is_pathable(command_basename):
        # If no message was passed, default this message.
        if exception_message is None:
            exception_message = (
                'Command "{}" not found in the current ${{PATH}} or '
                'found but not an executable file.'.format(command_basename))

        # Raise this exception.
        raise BetseCommandException(exception_message)

# ....................{ TESTERS                           }....................
@type_check
def is_pathable(command_basename: str) -> bool:
    '''
    ``True`` only if an external command with the passed basename exists (i.e.,
    corresponds to an executable file in the current ``${PATH}``).

    Caveats
    ----------
    For safety, avoid appending the passed basename by a platform-specific
    filetype -- especially, a Windows-specific filetype. On that platform, this
    function iteratively appends this basename by each filetype associated with
    executable files listed by the ``%PATHEXT%`` environment variable (e.g.,
    ``.bat``, ``.cmd``, ``.com``, ``.exe``) until the resulting basename is
    that of an executable file in the current ``%PATH%``.

    Parameters
    ----------
    command_basename : str
        Basename of the command to be searched for.

    Raises
    ----------
    BetsePathException
        If the passed string is *not* a basename (i.e., contains one or more
        directory separators).

    See Also
    ----------
    :func:`betse.util.os.command.cmds.is_command`
        A more general-purpose and hence generally useful tester returning
        whether a command with the passed path (regardless of whether that
        path is a basename) exists.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # If this string is *NOT* a pure basename, fail.
    pathnames.die_unless_basename(command_basename)

    # Return whether this command exists or not.
    return shutil.which(command_basename) is not None

# ....................{ GETTERS                           }....................
@type_check
def get_filename(command_basename: str) -> str:
    '''
    Absolute path of the command in the current ``${PATH}`` with the passed
    basename if found *or* raise an exception otherwise.

    Parameters
    ----------
    command_basename : str
        Basename of the command to return the absolute path of.

    Returns
    ----------
    str
        Absolute path of this command.

    Raises
    ----------
    BetseCommandException
        If no such command is found.
    '''

    # Absolute path of this command if found or "None" otherwise.
    command_filename = get_filename_or_none(command_basename)

    # If this command is *NOT* found, raise an exception.
    if command_filename is None:
        raise BetseCommandException(
            'Command "{}" not found.'.format(command_basename))

    # Return this path.
    return command_filename


@type_check
def get_filename_or_none(command_basename: str) -> StrOrNoneTypes:
    '''
    Absolute path of the command in the current ``${PATH}`` with the passed
    basename if found *or* ``None`` otherwise.

    Parameters
    ----------
    command_basename : str
        Basename of the command to return the absolute path of.

    Returns
    ----------
    StrOrNoneTypes
        Either;

        * Absolute path of this command if found.
        * ``None`` otherwise.
    '''

    # Avoid circular import dependencies.
    from betse.util.path import pathnames

    # If this string is *NOT* a pure basename, fail.
    pathnames.die_unless_basename(command_basename)

    # Return the absolute path of this command if found or "None" otherwise.
    return shutil.which(command_basename)

# ....................{ GETTERS ~ first                   }....................
@type_check
def get_first_basename(
    command_basenames: SequenceTypes, exception_message: str = None) -> str:
    '''
    First pathable string in the passed list (i.e., the first string that is
    the basename of a command in the current ``${PATH}``) if any *or* raise an
    exception otherwise.

    Parameters
    ----------
    command_basenames : SequenceTypes
        List of the basenames of all commands to be iteratively searched for
        (in descending order of preference).
    exception_message : optional[str]
        Optional exception message to be raised if no such string is pathable.
        Defaults to ``None``, in which case an exception message synthesized
        from the passed strings is raised.

    Returns
    ----------
    str
        First pathable string in the passed list.

    Raises
    ----------
    BetseCommandException
        If no passed strings are pathable.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text.string import strjoin

    # If this list contains the basename of a pathable command, return the
    # first such basename.
    for command_basename in command_basenames:
        if is_pathable(command_basename):
            return command_basename

    # Else, this list contains no such basename. Ergo, raise an exception.
    exception_message_suffix = (
        '{} not found in the current ${{PATH}}.'.format(
            strjoin.join_as_conjunction_double_quoted(*command_basenames)))

    # If a non-empty exception message is passed, suffix this message with this
    # detailed explanation.
    if exception_message:
        exception_message += ' ' + exception_message_suffix
    # Else, default this message to this detailed explanation.
    else:
        exception_message = exception_message_suffix

    # Raise this exception.
    raise BetseCommandException(exception_message)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level Numpy-based image facilities.

Design
----------
Most functions defined by this submodule are thin wrappers around similar
functionality provided by an arbitrary third-party dependency. As example, the
:func:`load_image` function:

* Currently wraps the :func:`Pillow.Image.load` method.
* Previously wrapped the :func:`scipy.misc.imread` function, which SciPy 1.0.0
  formally deprecated and SciPy 1.2.0 permanently removed. Indeed, SciPy 1.2.0
  broke backward compatibility with downstream consumers.

This submodule principally exists to mitigate the costs associated with similar
catastrophic upstream API changes in the future. We are ready this time.

Two principal choices exist:

* Pillow, the canonical image I/O library for Python. Pillow is a long-standing
  fork of PIL, whose stable API Pillow has inherited and carefully curated.
* imageio, a higher-level image I/O library for Python wrapping lower-level
  libraries (e.g., Pillow, FreeImage), whose API is required to generalize to
  multiple underlying libraries and is hence slightly less stable and
  feature-filled than Pillow. Moreover, imageio currently fails to support all
  fine-grained configurability required by this application (e.g., the capacity
  to convert images to grayscale).

Since the stable Pillow API provides all requisite features and the unstable
imageio API does *not*, this submodule is currently implemented in terms of
Pillow. Callers should *not* assume this to be the case, however.
'''

#FIXME: Revisit imageio when the following feature request is resolved in full:
#    https://github.com/imageio/imageio/issues/289

# ....................{ IMPORTS                           }....................
from PIL import Image
from betse.lib.numpy import nparray
from betse.util.io.log import logs
from betse.util.path import files, pathnames
from betse.util.type.types import type_check, NoneType, NumpyArrayType
from enum import Enum
from numpy import array

# ....................{ ENUMERATIONS                      }....................
#FIXME: Define the remainder of all Pillow modes. See the URL given below.
class ImageModeType(Enum):
    '''
    Enumeration of all types of **image modes** (i.e., :mod:`PIL`-specific
    identifier defining both the type and depth of all pixels in an image)
    supported by functions defined by this submodule.

    This enumeration wraps the non-typesafe, non-human-readable ``mode`` string
    parameter commonly accepted by :mod:`PIL` functions (e.g.,
    :func:`PIL.Image.new`) with typesafe, human-readable variants. The value of
    each member of this enumeration is the corresponding ``mode`` string.

    Attributes
    ----------
    COLOR_RGB : enum
        Integer-based true color mode *without* alpha transparency, providing
        8-bit red, green, and blue pixels.
    COLOR_RGBA : enum
        Integer-based true color mode *with* alpha transparency, providing
        8-bit red, green, blue, and alpha transparency pixels.
    GRAYSCALE_INT : enum
        Integer-based grayscale mode, providing 8-bit black and white pixels.
        Converting color images *not* already in this mode to this mode reduces
        the resulting images to grayscale according to the ITU-R 601-2 luma
        transform in a lossy integer manner (i.e., ``L = R * 299/1000 + G *
        587/1000 + B * 114/1000`` for each pixel such that the fractional
        portion of each such ``L`` is discarded).
    GRAYSCALE_FLOAT : enum
        Floating point-based grayscale mode, providing 32-bit floating point
        pixels. Converting color images *not* already in this mode to this mode
        reduces the resulting images to grayscale according to the ITU-R 601-2
        luma transform in a lossy floating point manner (i.e., ``L = R *
        299/1000 + G * 587/1000 + B * 114/1000`` for each pixel such that the
        fractional portion of each such ``L`` is preserved). This conversion
        generalizes the integer-based conversion to grayscale performed by
        converting to the :attr:`GRAYSCALE_INT` mode by preserving rather than
        discarding the fractional portion of each luma value. Note that the
        only file format currently supporting images of this mode is TIFF.

    See Also
    ----------
    http://pillow.readthedocs.io/en/3.4.x/handbook/concepts.html#modes
        Most recent official documentation for Pillow modes.
    '''

    COLOR_RGB       = 'RGB'
    COLOR_RGBA      = 'RGBA'
    GRAYSCALE_INT   = 'I'
    GRAYSCALE_FLOAT = 'F'

# ....................{ TYPES                             }....................
ImageModeOrNoneTypes = (ImageModeType, NoneType)
'''
Tuple of the type of all image mode enumeration members *and* of the singleton
``None`` object.
'''

# ....................{ CONVERTERS                        }....................
@type_check
def load_image(
    # Mandatory parameters.
    filename: str,

    # Optional parameters.
    is_signed: bool = True,
    mode: ImageModeOrNoneTypes = None,
) -> NumpyArrayType:
    '''
    Load the raw pixel data from the image with the passed filename into a
    multi-dimensional Numpy array and return this array.

    This array is guaranteed to be at least three-dimensional. Specifically, if
    this image is:

    * Greyscale, this array is three-dimensional such that:
      * The first dimension indexes each row of this image.
      * The second dimension indexes each column of the current row.
      * The third dimension indexes the grayscale value of the current pixel.
    * RGB or RGBA, this array is four-dimensional such that:
      * The first dimension indexes each row of this image.
      * The second dimension indexes each column of the current row.
      * The third dimension is either a 3-tuple ``(R, G, B)`` or 4-tuple
        ``(R, G, B, A)`` indexing each color component of the current pixel.
      * The fourth dimension indexes the value of the current color component.

    Caveats
    ----------
    When attempting to load user-defined images of arbitrary filetype, callers
    should pass the following parameters:

    * ``mode``.
    * ``is_signed`` to ``True``. Since this is (and *always* will be) the
      default, *not* passing this parameter satisfies this suggestion.

    Failure to do so invites subtle issues in computations falsely assuming the
    data type and shape of a returned array to be sane, which is *not* the case
    in common edge cases. Thanks to the heterogeneity of image file formats,
    returned arrays may exhibit anomalous features if any of these paremeters
    are *not* passed as suggested. In particular, the ``is_signed`` parameter
    should typically either *not* be passed or be passed as ``True``.

    Failure to do so instructs Pillow to produce an array with data type
    automatically corresponds to that of the input image. Since most (but *not*
    necessarily all) images reside in the :attr:`ImageModeType.COLOR_RGB` and
    :attr:`ImageModeType.COLOR_RGBA` colour spaces whose three- and
    four-channel pixel data is homogenously constrained onto unsigned bytes,
    most arrays returned by this function when explicitly passed an
    ``is_signed`` parameter of ``False`` will be **unsigned byte arrays**
    (i.e., arrays whose data types are :attr:`np.uint8`).

    Is that a subtle problem? **It is.**

    Python silently coerces scalar types as needed to preserve precision across
    operations that change precision. The canonical example is integer
    division.  In Python, dividing two integers that are *not* simple integer
    multiples of one another implicitly expands precision by producing a real
    number rather than integer (e.g., ``1 / 2 == 0.5`` rather than
    ``1 / 2 == 0``).

    On the other hand:

    * For all **signed Numpy arrays** (i.e., arrays whose data types are
      implicitly signed rather than explicitly unsigned), Numpy silently
      coerces the data types of these arrays as needed to preserve precision
      across precision-modifying operations.
    * For all **unsigned Numpy arrays** (i.e., arrays whose data types are
      explicitly unsigned rather than implicitly signed), Numpy silently
      preserves the unsigned facet of these arrays as needed by wrapping all
      numerical results to the integer range of these unsigned data types, thus
      discarding precision across precision-modifying operations.

    The canonical example is integer addition and substraction applied to
    unsigned byte arrays. Since unsigned bytes are confined to the integer
    range ``[0, 255]``, attempting to perform even seemingly trivial
    computation with unsigned byte arrays silently wraps results exceeding this
    range onto this range. The resulting arrays typically contain so-called
    "garbage data." As the following example shows, applying integer
    subtraction to signed but *not* unsigned Numpy arrays produces expected
    results:

        >>> import numpy as np
        >>> unsigned_garbage = np.array(((1,2), (3,4)), dtype=np.uint8)
        >>> unsigned_garbage[:,0] - unsigned_garbage[:,1]
        ... array([255, 255], dtype=uint8)
        >>> signed_nongarbage = np.array(((1,2), (3,4)))
        >>> signed_nongarbage[:,0] - signed_nongarbage[:,1]
        ... array([-1, -1])

    Design
    ----------
    This utility function is a thin wrapper around a similar function provided
    by some unspecified third-party dependency. This function currently wraps
    the :meth:`PIL.Image.open` method but previously wrapped the:

    * :func:`imageio.imread` function, which failed to expose support for
      colourspace conversion provided by Pillow.
    * :func:`scipy.misc.imread` function, which SciPy 1.0.0 formally deprecated
      and SciPy 1.2.0 permanently killed. Thus, SciPy 1.2.0 broke backward
      compatibility with downstream applications (notably, *this* application)
      requiring that API.

    This utility function principally exists to mitigate the costs associated
    with similar upstream API changes in the future. (We are ready this time.)

    Parameters
    ----------
    filename : str
        Absolute or relative filename of this image.
    is_signed : optional[bool]
        ``True`` only if converting the possibly unsigned array loaded from
        this image into a signed array. Defaults to ``True`` for the reasons
        detailed above. Since explicitly setting this to ``False`` invites
        errors in computations employing the returned array, callers should do
        so *ONLY* where these issues are acknowledged and handled
        appropriately.
    mode : ImageModeOrNoneTypes
        Type and depth of all pixels in the array loaded from this image,
        converted from this image's pixel data according to industry-standard
        image processing transforms implemented by :mod:`PIL`. Note that this
        is *not* the type and depth of all pixels in the input image, which
        :mod:`PIL` implicitly detects and hence requires no explicit
        designation. Defaults to ``None``, in which case no such conversion is
        performed (i.e., this image's pixel data is returned as is).

    Returns
    ----------
    ndarray
        Numpy array loaded from this image.
    '''

    # Log this load attempt.
    logs.log_debug('Loading image "%s"...', pathnames.get_basename(filename))

    # If this image does *NOT* exist, raise an exception.
    files.die_unless_file(filename)

    # In-memory image object loaded from this on-disk image file.
    image = Image.open(filename)

    # If the caller requests a mode conversion *AND* this image is not already
    # of the required mode, do this conversion.
    if mode is not None and mode != image.mode:
        image = image.convert(mode.value)

    # Numpy array converted from this image to be returned.
    image_array = array(image)

    # If converting unsigned to signed arrays, do so.
    if is_signed:
        image_array = nparray.to_signed(image_array)

    # Return this array.
    return image_array



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **data descriptor** (i.e., objects satisfying the data descriptor
protocol, typically defined at class scope) facilities.
'''

# ....................{ IMPORTS                           }....................
# from betse.util.type.types import type_check

# ....................{ CLASSES                           }....................
class DataDescriptorBound(object):
    '''
    High-level object encapsulating a low-level user-defined **data
    descriptor** (i.e., object satisfying the standard data descriptor
    protocol), bound at instantiation time to an instance of a user-defined
    class containing this descriptor and hence servicing exactly one instance.

    Motivation
    ----------
    This wrapper is principally intended to circumvent the conflict between
    immutable types in Python (e.g., :class:`int`, :class:`str`) and Python's
    pass-by-reference semantics.

    Consider two similar scenarios:

    * Passing an integer variable to a function as an integer parameter. In
      this case, modifying the latter does *not* modify the former.
    * Passing an instance of this class encapsulating an integer variable to a
      function as a parameter expecting only an instance of this class. In this
      case, modifying this instance actually modifies the integer variable
      encapsulated by this instance -- thereby circumventing this conflict.

    Design
    ----------
    Instances of this class access descriptors outside of class scope. Doing so
    is technically feasible but circumvents Python's default
    :meth:`object.__getattribute__` implementation transforming every attribute
    access ``b.x`` where ``x`` is a data descriptor declared at class scope
    into ``type(b).__dict__['x'].__get__(b, type(b))``. Circumventing this
    machinery typically defeats the purpose of instantiating a descriptor, but
    is warranted in this case.

    Why? Because this class neither requires nor desires this descriptor to
    ever be accessed via special machinery. Since the ``__get__()`` and
    ``__set__()`` methods defined by this descriptor already implement all
    functionality required by this class, explicitly calling these methods as
    is suffices.

    Attributes
    ----------
    data_desc : object
        Low-level data descriptor encapsulated by this higher-level object.
    _obj : object
        Object to be temporarily "bound" to this data descriptor for the
        duration of each call to the :meth:`get` and :meth:`set` methods.

    See Also
    ----------
    :class:`DataDescriptorUnbound`
        Data descriptor *not* bound to its parent object at instantiation time.
    :func:`betse.util.descriptor.expralias.expr_alias`
        Lower-level function creating data descriptors encapsulated by this
        higher-level class, intended to be used where declaring a descriptor at
        class scope is appropriate (e.g., where the data descriptor to be
        aliased is known at class definition time).
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, obj: object, data_desc: object) -> None:
        '''
        Encapsulate the passed data descriptor against the passed object.

        Parameters
        ----------
        obj : object
            Object to temporarily "bind" to this data descriptor for the
            duration of each call to the :meth:`get` and :meth:`set` methods.
        data_desc : None
            **Data descriptor** (i.e., object satisfying the standard data
            descriptor protocol) to be encapsulated.
        '''

        # Raise an exception unless this object is a data descriptor.
        die_unless_data_desc(data_desc)

        # Classify all passed parameters.
        self._obj = obj
        self.data_desc = data_desc

    # ..................{ GETTERS                           }..................
    def get(self) -> object:
        '''
        Value this data descriptor currently evaluates to.
        '''

        # Yes, the descriptor protocol explicitly supports direct method calls.
        # In particular, the "Invoking Descriptors" subsection of the official
        # "Descriptor HowTo Guide" notes:
        #
        #     A descriptor can be called directly by its method name. For
        #     example, d.__get__(obj).
        #
        # Since this data descriptor is accessed with respect to an instance
        # variable rather than a class, the second parameter passed to this
        # method call is "None".
        #
        # See: https://docs.python.org/3/howto/descriptor.html
        return self.data_desc.__get__(self._obj, None)

    # ..................{ SETTERS                           }..................
    def set(self, value: object) -> None:
        '''
        Set this data descriptor's current value to the passed value.
        '''

        self.data_desc.__set__(self._obj, value)


class DataDescriptorUnbound(object):
    '''
    High-level object encapsulating a low-level user-defined **data
    descriptor** (i.e., object satisfying the standard data descriptor
    protocol), *not* bound at instantiation time to an instance of a
    user-defined class containing this descriptor and hence servicing multiple
    different instances.

    Attributes
    ----------
    data_desc : object
        Low-level data descriptor encapsulated by this higher-level object.

    See Also
    ----------
    :class:`DataDescriptorBound`
        Data descriptor bound to its parent object at instantiation time.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, data_desc: object) -> None:
        '''
        Encapsulate the passed data descriptor.

        Parameters
        ----------
        data_desc : None
            **Data descriptor** (i.e., object satisfying the standard data
            descriptor protocol) to be encapsulated.
        '''

        # Raise an exception unless this object is a data descriptor.
        die_unless_data_desc(data_desc)

        # Classify all passed parameters.
        self.data_desc = data_desc

    # ..................{ GETTERS                           }..................
    def get(self, obj: object) -> object:
        '''
        Value this data descriptor currently evaluates to.

        Parameters
        ----------
        obj : object
            Object to be temporarily bound to this data descriptor for the
            duration of this call.

        Returns
        ----------
        object
            This data descriptor's current value.
        '''

        return self.data_desc.__get__(obj, None)

    # ..................{ SETTERS                           }..................
    def set(self, obj: object, value: object) -> None:
        '''
        Set this data descriptor's current value to the passed value.

        Parameters
        ----------
        obj : object
            Object to be temporarily bound to this data descriptor for the
            duration of this call.
        value : object
            Value to set this data descriptor to.
        '''

        self.data_desc.__set__(obj, value)

# ....................{ EXCEPTIONS                        }....................
def die_unless_data_desc(data_desc: object) -> None:
    '''
    Raise an exception unless the passed object is a **data descriptor** (i.e.,
    object defining both the special ``__get__`` and ``__set__`` methods and
    hence satisfying the standard data descriptor protocol).

    Parameters
    ----------
    data_desc : object
        Object to be tested.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.obj import objtest

    # Raise an exception unless all special methods required by the standard
    # data descriptor protocol are bound to this object.
    objtest.die_unless_has_method(data_desc, '__get__', '__set__')
